from typing import Sequence
import random

import torch
from torch import Tensor
import torchvision.transforms as T  # noqa: N812

from hearth.modules import BaseModule
from hearth.vision.ops import square_center_crop


class NormalizeImg(BaseModule):
    """normalize a 3 channel image or a batch of 3 channel images.

    Inputs are expected to be floats in range 0.0:1.0 of shape (B?, C, W, H) where B? is an
    optional batch dimension.

    Args:
        mean: mean for each channel, defaults to imagenet defaults as outlined in torchvision.
        std: std for each channel, defaults to imagenet defaults as outlined in torchvision.

    Example:
        >>> from hearth.vision.transforms import NormalizeImg
        >>> _ = torch.manual_seed(0)
        >>>
        >>> # use torchvision imagenet defaults...
        >>> normalize = NormalizeImg()
        >>> single_image = torch.rand(3, 64, 64)
        >>> normalized_single_image = normalize(single_image)
        >>> normalized_single_image.mean((1, 2))
        tensor([0.0448, 0.2087, 0.4471])

        >>> normalized_single_image.std((1, 2))
        tensor([1.2547, 1.2854, 1.2874])

        >>> batch_images = torch.rand(5, 3, 64, 64)
        >>> normalized_batch_images = normalize(batch_images)
        >>> normalized_batch_images.mean((0, 2, 3))
        tensor([0.0778, 0.1930, 0.4113])

        >>> normalized_batch_images.std((0, 2, 3))
        tensor([1.2605, 1.2804, 1.2793])
    """

    def __init__(
        self,
        mean: Sequence[float] = (0.485, 0.456, 0.406),
        std: Sequence[float] = (0.229, 0.224, 0.225),
    ):
        super().__init__()
        self.register_buffer('mean', torch.tensor(mean).reshape(-1, 1, 1))
        self.register_buffer('std', torch.tensor(std).reshape(-1, 1, 1))

    def forward(self, x: Tensor) -> Tensor:
        return (x - self.mean) / self.std


class ResizeCrop(BaseModule):
    """resize a single image by scaling smallest dimension and center cropping with optional shift.

    Args:
        size: new size for square image.
        noise: for shifting the crop window along the longest dimension Defaults to 0.0 (no noise).

    Example:
        >>> import torch
        >>> from hearth.vision.transforms import ResizeCrop
        >>>
        >>> resize = ResizeCrop(224, crop_noise=0.5)
        >>> img = torch.randint(0, 255, size=(3, 1024, 986))
        >>> resize(img).shape
        torch.Size([3, 224, 224])
    """

    def __init__(self, size: int, crop_noise: float = 0.0):
        super().__init__()
        self.size = size
        self.crop_noise = crop_noise
        self.resize = T.Resize(self.size)

    def forward(self, img: torch.Tensor) -> torch.Tensor:
        return square_center_crop(self.resize(img), noise=self.crop_noise)


class RandomFlipOrient(BaseModule):
    """randomly flip orientation of a single image

    Args:
        p: probability of flipping orientaion. Defaults to 0.5.

    Example:
        >>> from hearth.vision.transforms import RandomFlipOrient
        >>>
        >>> # set p=1.0 to ensure stuff happens
        >>> transform = RandomFlipOrient(p=1.0)
        >>> img = torch.randint(0, 255, size=(3, 128, 256))
        >>> transform(img).shape
        torch.Size([3, 256, 128])
    """

    def __init__(self, p: float = 0.5):
        super().__init__()
        self.p = p

    def forward(self, img: torch.Tensor) -> torch.Tensor:
        if random.random() <= self.p:
            return img.swapaxes(-2, -1).contiguous()
        return img



"""only needed post torch 1.10 updates patches improvements that allow us to handle tensordicts
"""

import torch
import re
import collections
from torch._six import string_classes


np_str_obj_array_pattern = re.compile(r'[SaUO]')


default_collate_err_msg_format = (
    "default_collate: batch must contain tensors, numpy arrays, numbers, "
    "dicts or lists; found {}"
)


def default_collate(batch):  # noqa: C901
    """Function that takes in a batch of data and puts the elements within the batch
    into a tensor with an additional outer dimension - batch size. The exact output type can be
    a :class:`torch.Tensor`, a `Sequence` of :class:`torch.Tensor`, a
    Collection of :class:`torch.Tensor`, or left unchanged, depending on the input type.
    This is used as the default function for collation when
    `batch_size` or `batch_sampler` is defined in :class:`~torch.utils.data.DataLoader`.

    Args:
        batch: a single batch to be collated
    Examples:
        >>> from collections import namedtuple
        >>> # Example with a batch of `int`s:
        >>> default_collate([0, 1, 2, 3])
        tensor([0, 1, 2, 3])
        >>> # Example with a batch of `str`s:
        >>> default_collate(['a', 'b', 'c'])
        ['a', 'b', 'c']
        >>> # Example with `Map` inside the batch:
        >>> default_collate([{'A': 0, 'B': 1}, {'A': 100, 'B': 100}])
        {'A': tensor([  0, 100]), 'B': tensor([  1, 100])}
        >>> # Example with `NamedTuple` inside the batch:
        >>> Point = namedtuple('Point', ['x', 'y'])
        >>> default_collate([Point(0, 0), Point(1, 1)])
        Point(x=tensor([0, 1]), y=tensor([0, 1]))
        >>> # Example with `Tuple` inside the batch:
        >>> default_collate([(0, 1), (2, 3)])
        [tensor([0, 2]), tensor([1, 3])]
        >>> # Example with `List` inside the batch:
        >>> default_collate([[0, 1], [2, 3]])
        [tensor([0, 2]), tensor([1, 3])]
    """
    elem = batch[0]
    elem_type = type(elem)
    if isinstance(elem, torch.Tensor):
        out = None
        if torch.utils.data.get_worker_info() is not None:
            # If we're in a background process, concatenate directly into a
            # shared memory tensor to avoid an extra copy
            numel = sum(x.numel() for x in batch)
            storage = elem.storage()._new_shared(numel)
            out = elem.new(storage)
        return torch.stack(batch, 0, out=out)
    elif (
        elem_type.__module__ == 'numpy'
        and elem_type.__name__ != 'str_'
        and elem_type.__name__ != 'string_'
    ):
        if elem_type.__name__ in ['ndarray', 'memmap']:
            # array of string classes and object
            if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                raise TypeError(default_collate_err_msg_format.format(elem.dtype))

            return default_collate([torch.as_tensor(b) for b in batch])
        elif elem.shape == ():  # scalars
            return torch.as_tensor(batch)
    elif isinstance(elem, float):
        return torch.tensor(batch, dtype=torch.float64)
    elif isinstance(elem, int):
        return torch.tensor(batch)
    elif isinstance(elem, string_classes):
        return batch
    elif isinstance(elem, collections.abc.Mapping):
        return elem_type({key: default_collate([d[key] for d in batch]) for key in elem})

    elif isinstance(elem, tuple) and hasattr(elem, '_fields'):  # namedtuple
        return elem_type(*(default_collate(samples) for samples in zip(*batch)))
    elif isinstance(elem, collections.abc.Sequence):
        # check to make sure that the elements in batch have consistent size
        it = iter(batch)
        elem_size = len(next(it))
        if any(len(elem) != elem_size for elem in it):
            raise RuntimeError('each element in list of batch should be of equal size')
        transposed = list(zip(*batch))  # It may be accessed twice, so we use a list.

        if isinstance(elem, tuple):
            return [default_collate(samples) for samples in transposed]  # Backwards compatibility.
        try:
            return elem_type([default_collate(samples) for samples in transposed])
        except TypeError:
            # The sequence type may not support `__init__(iterable)` (e.g., `range`).
            return [default_collate(samples) for samples in transposed]

    raise TypeError(default_collate_err_msg_format.format(elem_type))



from typing import Optional, Tuple
import math

import torch
from torch import nn
from hearth.modules.base import BaseModule
from hearth.modules.wrappers import TimeMasked


class SelfAttention(BaseModule):
    """Scaled self attention layer for timeseries.

    Args:
        in_features: number of input features.
        out_features: number of output features, if None defaults to in_features. Defaults to None.
        n_heads: number of attention heads, must evenly divide out_features. Defaults to 1.
        bias: include bias in all layers. Defaults to True.
        dropout: attention dropout rate. Defaults to 0.1.
        drop_head: If true drop attention from entire attention heads as dropout strategy otherwise
            drop timesteps from different heads. Defaults to False.

    Example:
        >>> from torch import nn
        >>> from hearth.modules import SelfAttention
        >>>
        >>> layer = SelfAttention(16, 32, n_heads=4)
        >>> mask = torch.tensor([[ True,  True,  True, False, False],
        ...                      [ True,  True,  True,  True,  True]])
        >>> inp = torch.rand(2, 5, 16)
        >>> out = layer(inp, mask)
        >>> out.shape
        torch.Size([2, 5, 32])

        >>> (out == 0.0).all(-1)
        tensor([[False, False, False,  True,  True],
               [False, False, False, False, False]])

    """

    def __init__(
        self,
        in_features: int,
        out_features: Optional[int] = None,
        n_heads: int = 1,
        bias: bool = True,
        dropout=0.1,
        drop_head: bool = False,
    ):
        super().__init__()

        self.in_features = in_features
        self.out_features = out_features or in_features
        if self.out_features % n_heads != 0:
            raise ValueError('out_features must be divisible by n_heads')
        self.n_heads = n_heads
        self.dim_per_head = int(self.out_features // self.n_heads)
        # stack the whole goddamn thing together...
        self.qkv = TimeMasked(nn.Linear(self.in_features, 3 * self.out_features, bias=bias))
        self.scale = math.sqrt(self.dim_per_head)
        self.dropout = nn.Dropout2d(dropout) if drop_head else nn.Dropout(dropout)
        self.output = TimeMasked(nn.Linear(self.out_features, self.out_features, bias=bias))

    def _shape(self, x: torch.Tensor) -> torch.Tensor:
        return x.view(x.shape[0], -1, self.n_heads, self.dim_per_head).transpose(1, 2)

    def _unshape(self, x: torch.Tensor) -> torch.Tensor:
        return x.transpose(1, 2).contiguous().view(x.shape[0], -1, self.out_features)

    def _get_qkv(
        self, x: torch.Tensor, mask: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        q, k, v = self.qkv(x, mask).chunk(3, -1)
        return self._shape(q), self._shape(k), self._shape(v)

    def _compute_attn(self, q, k, mask):
        b = mask.shape[0]
        scores = torch.matmul(q, k.transpose(2, 3))  # (bs, n_heads, q_length, k_length)
        padding_mask = (~mask).view((b, 1, 1, -1)).expand_as(scores)
        weights = nn.functional.softmax(scores, dim=-1)  # (bs, n_heads, q_length, k_length)
        weights = weights.masked_fill(
            padding_mask.transpose(-1, 2), 0.0
        )  # (bs, n_heads, q_length, k_length)
        return weights

    def forward(self, x: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        q, k, v = self._get_qkv(x, mask)
        q = q / self.scale
        attn = self.dropout(self._compute_attn(q, k, mask))
        context = torch.matmul(attn, v)  # (bs, n_heads, q_length, dim_per_head)
        return self.output(self._unshape(context), mask)



from .attention import SelfAttention
from .base import BaseModule
from .wrappers import Residual, ReZero, TimeMasked
from .normalization import LayerNormSimple, MaskedLayerNorm
from .embeddings import AbsolutePositionalEmbedding, TransformerEmbedding
from .transformer import Boom, TransformerEncoder, TransformerEncoderBlock

__all__ = [
    'BaseModule',
    'Residual',
    'ReZero',
    'LayerNormSimple',
    'MaskedLayerNorm',
    'AbsolutePositionalEmbedding',
    'TimeMasked',
    'SelfAttention',
    'Boom',
    'TransformerEncoder',
    'TransformerEncoderBlock',
    'TransformerEmbedding',
]



import torch
from torch import nn

from hearth.modules import BaseModule
from hearth.modules.normalization import MaskedLayerNorm


class AbsolutePositionalEmbedding(BaseModule):
    """Absolute learned positional embeddings a la bert.

    Args:
        features: number of embedding features.
        max_len: max sequence length.
        padding_idx: used to mask padding timesteps

    Example:
        >>> from hearth.modules import AbsolutePositionalEmbedding
        >>>
        >>> emb = AbsolutePositionalEmbedding(256, max_len=512)
        >>> tokens = torch.tensor([[99, 6, 55, 1, 0, 0],
        ...                        [8, 22, 7, 8, 3, 11]])
        >>> out = emb(tokens)
        >>> out.shape
        torch.Size([2, 6, 256])

        >>> (out[tokens == 0] == 0).all()
        tensor(True)
    """

    def __init__(self, features: int, max_len: int = 512, padding_idx: int = 0):
        super().__init__()
        self.out_features = features
        self.max_len = max_len
        self.padding_idx = padding_idx
        self.embedding = nn.Embedding(self.max_len + 1, features, padding_idx=self.padding_idx)
        self.register_buffer(
            'position_ids', torch.arange(1, self.max_len + 1).expand((1, -1)), persistent=False
        )

    def forward(self, tokens: torch.Tensor) -> torch.Tensor:
        timesteps = tokens.size(1)

        position_ids = self.position_ids[:, :timesteps]  # type: ignore
        position_ids = position_ids.expand_as(tokens).masked_fill(
            (tokens == self.padding_idx), 0
        )  # (bs, max_seq_length)

        return self.embedding(position_ids)


class TransformerEmbedding(BaseModule):
    """simple version of embedding used in transformer models.

    Note:
      this embedding does not use token type embeddings as used in some bert models. If
      using with a pretrained model it's reccomended that you first add these to word embeddings.

    Args:
        vocab_size: vocabulary size for word embeddings.
        features: number of features in embedding space.
        max_len: maximum sequence length for positional embeddings. Defaults to 512.
        dropout: dropout rate. Defaults to 0.1.
        layer_norm_eps: epsilon for layer normalization. Defaults to 1e-12.
        padding_idx: index for non-valid padding timesteps. Defaults to 0.

    Example:
        >>> from torch import nn
        >>> from hearth.modules import TransformerEmbedding
        >>>
        >>> emb = TransformerEmbedding(1000, 256, padding_idx=0)

        >>> tokens = torch.tensor([[99, 6, 55, 1, 0, 0],
        ...                        [8, 22, 7, 8, 3, 11]])
        >>> out = emb(tokens)
        >>> out.shape
        torch.Size([2, 6, 256])

        >>> (out == 0.0).all(-1)
        tensor([[False, False, False, False, True,  True],
                [False, False, False, False, False, False]])

        >>> emb.build_mask(tokens)
        tensor([[ True,  True,  True,  True, False, False],
                [ True,  True,  True,  True,  True,  True]])
    """

    def __init__(
        self,
        vocab_size: int,
        features: int,
        max_len: int = 512,
        dropout: float = 0.1,
        layer_norm_eps: float = 1e-12,
        padding_idx: int = 0,
    ):
        super().__init__()
        self.out_features = features
        self.padding_idx = padding_idx
        self.word_embeddings = nn.Embedding(vocab_size, features, padding_idx=padding_idx)
        self.position_embeddings = AbsolutePositionalEmbedding(
            features, max_len, padding_idx=padding_idx
        )
        self.norm = MaskedLayerNorm(features, eps=layer_norm_eps)
        self.dropout = nn.Dropout(dropout)

    @torch.jit.export
    def build_mask(self, tokens: torch.Tensor) -> torch.Tensor:
        """get a mask where all valid timesteps are True"""
        return tokens != self.padding_idx

    def forward(self, tokens: torch.Tensor) -> torch.Tensor:
        x = self.word_embeddings(tokens) + self.position_embeddings(tokens)
        x = self.norm(x, self.build_mask(tokens))
        return self.dropout(x)



from dataclasses import dataclass
from hearth.metrics.mixins import (
    HardBinaryMixin,
    MaskingMixin,
    AccuracyMixin,
    ArgmaxMixin,
    RecallMixin,
    PrecisionMixin,
    F1Mixin,
    FBetaMixin,
    BinaryMixin,
    OneHotMixin,
    AverageMixin,
)


@dataclass
class BinaryAccuracy(HardBinaryMixin, MaskingMixin, AccuracyMixin):
    """binary accruracy over possibly unnormalized values (see ``from_logits``) with target masking\
     support.

    Args:
        mask_target: mask targets equal to this value. defaults to ``-1``.
        from_logits: if ``True`` inputs are expected to be unnormalized and a sigmoid
            function will be applied before comparison to targets. defaults to ``False``.

    Example:
        >>> import torch
        >>> _ = torch.manual_seed(0)
        >>> from hearth.metrics import BinaryAccuracy
        >>>
        >>>
        >>> metric = BinaryAccuracy()
        >>> metric
        BinaryAccuracy(mask_target=-1, from_logits=False)

        by default ``from_logits`` is false so we expect inputs to be normalized.

        >>> targets = torch.randint(1, size=(100,)) # (batch, 1)
        >>> predictions = torch.normal(0, 1, size=(100, 1)) # (batch, 1)
        >>>
        >>> metric(predictions.sigmoid(), targets)
        tensor(0.4400)

        alternatively we could specify ``from_logits=True`` in which case inputs
        will be sigmoided for us before comparing with targets.

        >>> metric = BinaryAccuracy(from_logits=True)
        >>> metric(predictions, targets) # no sigmoid!
        tensor(0.4400)

        this metric also supports masked targets in the case of variable length sequence data
        which should be **batch first** ``(batch, time)`` or ``(batch, time, 1)``.

        >>> metric = BinaryAccuracy()
        >>> targets = torch.tensor([[1, 1, 1, -1], [1, 1, -1, -1], [1, 1, 1, 0]]) # (batch, time)
        >>> predictions = torch.ones(3, 4)
        >>> metric(predictions, targets)
        tensor(0.8889)
    """

    pass


@dataclass
class CategoricalAccuracy(MaskingMixin, ArgmaxMixin, AccuracyMixin):
    """categorical accuracy over possibly unnormalized scores given target indices.

    Built to have a similar interface to\
     `nn.CrossEntropyLoss <https://pytorch.org/docs/stable/data.html#torch.nn.CrossEntropyLoss>`_.

    Args:
        mask_target: mask this value if seen in the targets to this value. defaults to ``-1``

    Example:
        >>> import torch
        >>> _ = torch.manual_seed(0)
        >>> from hearth.metrics import CategoricalAccuracy
        >>>
        >>> metric = CategoricalAccuracy()
        >>> metric
        CategoricalAccuracy(mask_target=-1)


        >>> targets = torch.randint(4, size=(100,)) # (bath,)
        >>> predictions = torch.normal(0, 1, size=(100, 4)) # (batch, classes)
        >>> metric(predictions, targets)
        tensor(0.2200)


        mask targets using the ``mask_target``, particularly useful for mixed length
        sequence predicions....

        >>> targets = torch.tensor([[0, 5, 9, -1], [2, 3, -1, -1], [1, 6, 3, 4]]) # (batch, time)
        >>> predictions = torch.normal(0, 1, size=(3, 4,  9)) # (batch, time, classes)
        >>> metric(predictions, targets)
        tensor(0.1111)
    """

    pass


@dataclass
class BinaryRecall(HardBinaryMixin, MaskingMixin, RecallMixin):
    """"binary recall over possibly un-normalized values (see ``from_logits``) with target masking\
     support.

    Args:
        mask_target: mask targets equal to this value. defaults to ``-1``.
        from_logits: if ``True`` inputs are expected to be unnormalized and a sigmoid
            function will be applied before comparison to targets. defaults to ``False``.
    Example:
        >>> import torch
        >>> from hearth.metrics import BinaryRecall
        >>>
        >>> inputs = torch.tensor([0.7116, 0.6470, 0.5039, 0.9953, 0.8948, 0.4229, 0.8654, 0.8108])
        >>> targets = torch.tensor([0., 1., 1., 1., 0., 1., 0., 0.])
        >>>
        >>> recall = BinaryRecall()
        >>> recall(inputs, targets)
        tensor(0.7500)

        works fine with an extra dim:

        >>> recall(inputs.unsqueeze(-1), targets.unsqueeze(-1))
        tensor(0.7500)

        use the ``from_logits`` option if your inputs will not be sigmoid squashed:

        >>> recall = BinaryRecall(from_logits=True)
        >>> unsigmoided = torch.log(inputs/(1-inputs))
        >>> recall(unsigmoided, targets)
        tensor(0.7500)

        for cases like variable lenght time sequences use the mask target option:

        >>> recall = BinaryRecall(mask_target=-1)
        >>> inputs = inputs.reshape(2, 4) # (batch, sequence length)
        >>> # note we mask some timesteps with -1
        >>> targets = torch.tensor([[0., 1., 1., -1],
        ...                         [0., 1, -1, -1]])
        >>> recall(inputs, targets)
        tensor(0.6667)

    """

    pass


@dataclass
class BinaryPrecision(HardBinaryMixin, MaskingMixin, PrecisionMixin):
    """"binary precision over possibly un-normalized values (see ``from_logits``) with target \
    masking support.

    Args:
        mask_target: mask targets equal to this value. defaults to ``-1``.
        from_logits: if ``True`` inputs are expected to be unnormalized and a sigmoid
            function will be applied before comparison to targets. defaults to ``False``.
    Example:
        >>> import torch
        >>> from hearth.metrics import BinaryPrecision
        >>>
        >>> inputs = torch.tensor([0.7116, 0.6470, 0.5039, 0.9953, 0.8948, 0.4229, 0.8654, 0.8108])
        >>> targets = torch.tensor([0., 1., 1., 1., 0., 1., 0., 0.])
        >>>
        >>> precision = BinaryPrecision()
        >>> precision(inputs, targets)
        tensor(0.4286)

        works fine with an extra dim:

        >>> precision(inputs.unsqueeze(-1), targets.unsqueeze(-1))
        tensor(0.4286)

        use the ``from_logits`` option if your inputs will not be sigmoid squashed:

        >>> precision = BinaryPrecision(from_logits=True)
        >>> unsigmoided = torch.log(inputs/(1-inputs))
        >>> precision(unsigmoided, targets)
        tensor(0.4286)

        for cases like variable lenght time sequences use the mask target option:

        >>> precision = BinaryPrecision(mask_target=-1)
        >>> inputs = inputs.reshape(2, 4) # (batch, sequence length)
        >>> # note we mask some timesteps with -1
        >>> targets = torch.tensor([[0., 1., 1., -1],
        ...                         [0., 1, -1, -1]])
        >>> precision(inputs, targets)
        tensor(0.5000)
    """

    pass


@dataclass
class SoftBinaryRecall(BinaryMixin, MaskingMixin, RecallMixin):
    """a soft version of binary recall over possibly un-normalized values (see ``from_logits``)\
     with target masking support.

    Args:
        mask_target: mask targets equal to this value. defaults to ``-1``.
        from_logits: if ``True`` inputs are expected to be unnormalized and a sigmoid
            function will be applied before comparison to targets. defaults to ``False``.

    Example:
        >>> import torch
        >>> from hearth.metrics import SoftBinaryRecall
        >>>
        >>> inputs = torch.tensor([0.7116, 0.6470, 0.5039, 0.9953, 0.8948, 0.4229, 0.8654, 0.8108])
        >>> targets = torch.tensor([0., 1., 1., 1., 0., 1., 0., 0.])
        >>>
        >>> softrecall = SoftBinaryRecall()
        >>> softrecall(inputs, targets)
        tensor(0.6423)

        works fine with an extra dim:

        >>> softrecall(inputs.unsqueeze(-1), targets.unsqueeze(-1))
        tensor(0.6423)

        use the ``from_logits`` option if your inputs will not be sigmoid squashed:

        >>> softrecall = SoftBinaryRecall(from_logits=True)
        >>> unsigmoided = torch.log(inputs/(1-inputs))
        >>> softrecall(unsigmoided, targets)
        tensor(0.6423)

        for cases like variable lenght time sequences use the mask target option:

        >>> softrecall = SoftBinaryRecall(mask_target=-1)
        >>> inputs = inputs.reshape(2, 4) # (batch, sequence length)
        >>> # note we mask some timesteps with -1
        >>> targets = torch.tensor([[0., 1., 1., -1],
        ...                         [0., 1, -1, -1]])
        >>> softrecall(inputs, targets)
        tensor(0.5246)
    """

    pass


@dataclass
class SoftBinaryPrecision(BinaryMixin, MaskingMixin, PrecisionMixin):
    """a soft version of binary precision over possibly un-normalized values (see ``from_logits``)\
     with target masking support.

    Args:
        mask_target: mask targets equal to this value. defaults to ``-1``.
        from_logits: if ``True`` inputs are expected to be unnormalized and a sigmoid
            function will be applied before comparison to targets. defaults to ``False``.

    Example:
        >>> import torch
        >>> from hearth.metrics import SoftBinaryPrecision
        >>>
        >>> inputs = torch.tensor([0.7116, 0.6470, 0.5039, 0.9953, 0.8948, 0.4229, 0.8654, 0.8108])
        >>> targets = torch.tensor([0., 1., 1., 1., 0., 1., 0., 0.])
        >>>
        >>> softprecision = SoftBinaryPrecision()
        >>> softprecision(inputs, targets)
        tensor(0.4390)

        works fine with an extra dim:

        >>> softprecision(inputs.unsqueeze(-1), targets.unsqueeze(-1))
        tensor(0.4390)

        use the ``from_logits`` option if your inputs will not be sigmoid squashed:

        >>> softprecision = SoftBinaryPrecision(from_logits=True)
        >>> unsigmoided = torch.log(inputs/(1-inputs))
        >>> softprecision(unsigmoided, targets)
        tensor(0.4390)

        for cases like variable lenght time sequences use the mask target option:

        >>> softprecision = SoftBinaryPrecision(mask_target=-1)
        >>> inputs = inputs.reshape(2, 4) # (batch, sequence length)
        >>> # note we mask some timesteps with -1
        >>> targets = torch.tensor([[0., 1., 1., -1],
        ...                         [0., 1, -1, -1]])
        >>> softprecision(inputs, targets)
        tensor(0.4949)
    """

    pass


@dataclass
class BinaryF1(HardBinaryMixin, MaskingMixin, F1Mixin):
    """binary f1 score over possibly un-normalized values (see ``from_logits``) with target \
    masking support.

    Args:
        mask_target: mask targets equal to this value. defaults to ``-1``.
        from_logits: if ``True`` inputs are expected to be unnormalized and a sigmoid
            function will be applied before comparison to targets. defaults to ``False``.

    Example:
        >>> import torch
        >>> from hearth.metrics import BinaryF1
        >>>
        >>> inputs = torch.tensor([0.7116, 0.6470, 0.5039, 0.9953, 0.8948, 0.4229, 0.8654, 0.8108])
        >>> targets = torch.tensor([0., 1., 1., 1., 0., 1., 0., 0.])
        >>>
        >>> f1 = BinaryF1()
        >>> f1(inputs, targets)
        tensor(0.5455)

        works fine with an extra dim:

        >>> f1(inputs.unsqueeze(-1), targets.unsqueeze(-1))
        tensor(0.5455)

        use the ``from_logits`` option if your inputs will not be sigmoid squashed:

        >>> f1 = BinaryF1(from_logits=True)
        >>> unsigmoided = torch.log(inputs/(1-inputs))
        >>> f1(unsigmoided, targets)
        tensor(0.5455)

        for cases like variable lenght time sequences use the mask target option:

        >>> f1 = BinaryF1(mask_target=-1)
        >>> inputs = inputs.reshape(2, 4) # (batch, sequence length)
        >>> # note we mask some timesteps with -1
        >>> targets = torch.tensor([[0., 1., 1., -1],
        ...                         [0., 1, -1, -1]])
        >>> f1(inputs, targets)
        tensor(0.5714)
    """

    pass


@dataclass
class BinaryFBeta(HardBinaryMixin, MaskingMixin, FBetaMixin):
    """binary fbeta score over possibly un-normalized values (see ``from_logits``) with target \
    masking support.

    Args:
        beta: beta value for weighting precision and recall. ``beta < 1`` weights precision higher,
            while ``beta > 1`` gives more weight to recall. Defaults to 1 (making the default the
             same as f1 score).
        mask_target: mask targets equal to this value. defaults to ``-1``.
        from_logits: if ``True`` inputs are expected to be unnormalized and a sigmoid
            function will be applied before comparison to targets. defaults to ``False``.

    Example:
        >>> import torch
        >>> from hearth.metrics import BinaryFBeta
        >>>
        >>> inputs = torch.tensor([0.7116, 0.6470, 0.5039, 0.9953, 0.8948, 0.4229, 0.8654, 0.8108])
        >>> targets = torch.tensor([0., 1., 1., 1., 0., 1., 0., 0.])
        >>>
        >>> fbeta = BinaryFBeta(beta=.5)
        >>> fbeta(inputs, targets)
        tensor(0.4687)

        works fine with an extra dim:

        >>> fbeta(inputs.unsqueeze(-1), targets.unsqueeze(-1))
        tensor(0.4687)

        use the ``from_logits`` option if your inputs will not be sigmoid squashed:

        >>> fbeta = BinaryFBeta(beta=.5, from_logits=True)
        >>> unsigmoided = torch.log(inputs/(1-inputs))
        >>> fbeta(unsigmoided, targets)
        tensor(0.4687)

        for cases like variable lenght time sequences use the mask target option:

        >>> fbeta = BinaryFBeta(beta=.5, mask_target=-1)
        >>> inputs = inputs.reshape(2, 4) # (batch, sequence length)
        >>> # note we mask some timesteps with -1
        >>> targets = torch.tensor([[0., 1., 1., -1],
        ...                         [0., 1, -1, -1]])
        >>> fbeta(inputs, targets)
        tensor(0.5263)
    """

    pass


@dataclass
class CategoricalRecall(MaskingMixin, OneHotMixin, RecallMixin, AverageMixin):
    """categorical recall over possibly unnormalized scores given target indices.

    Built to have a somewhat similar interface to\
     `nn.CrossEntropyLoss <https://pytorch.org/docs/stable/data.html#torch.nn.CrossEntropyLoss>`_.

    Args:
        mask_target: mask targets equal to this value. defaults to ``-1``.

    Example:
        >>> import torch
        >>> from hearth.metrics import CategoricalRecall
        >>>
        >>> # here inputs is batch, feats and is not nessisarily normalized
        >>> inputs = torch.tensor([[-0.2956,  1.6050,  0.4113, -1.9041],
        ...                        [ 0.2095,  1.2959, -1.2466,  2.2302],
        ...                        [ 0.4702, -0.7506,  1.6751,  0.3370],
        ...                        [-0.4504,  0.5301, -1.1206, -0.5896],
        ...                        [ 0.7439,  0.4022,  0.5913,  0.1511],
        ...                        [-0.0523, -1.0082,  0.5536, -1.2748],
        ...                        [ 0.5151, -0.9396,  0.7223, -0.5500],
        ...                        [ 0.1083,  2.7311,  1.4429,  1.0640]])
        >>> targets = torch.tensor([3, 1, 1, 2, 0, 3, 0, 2]) # true class indices
        >>>
        >>> recall = CategoricalRecall()
        >>> recall(inputs, targets)
        tensor(0.1250)

        if you're predicting categories for variable length sequences
        you can mask your targets with ``mask_target`` (-1 by default)
        predictions at those timesteps will then be excluded from the metric.

        >>> sequence_inputs = inputs.reshape(2, 4, 4) # (batch, timesteps, classes)
        >>> # note the last two timesteps in targets are masked with -1
        >>> masked_targets = torch.tensor([[ 3,  1,  1, 2],
        ...                                [ 0, 3, -1, -1]]) # (batch, timesteps)
        >>> recall(sequence_inputs, masked_targets)
        tensor(0.2500)
    """

    pass


@dataclass
class CategoricalPrecision(MaskingMixin, OneHotMixin, PrecisionMixin, AverageMixin):
    """categorical precision over possibly unnormalized scores given target indices.

    Built to have a somewhat similar interface to\
     `nn.CrossEntropyLoss <https://pytorch.org/docs/stable/data.html#torch.nn.CrossEntropyLoss>`_.


    Args:
        mask_target: mask targets equal to this value. defaults to ``-1``.

    Example:
        >>> import torch
        >>> from hearth.metrics import CategoricalPrecision
        >>>
        >>> # here inputs is batch, feats and is not nessisarily normalized
        >>> inputs = torch.tensor([[-0.2956,  1.6050,  0.4113, -1.9041],
        ...                        [ 0.2095,  1.2959, -1.2466,  2.2302],
        ...                        [ 0.4702, -0.7506,  1.6751,  0.3370],
        ...                        [-0.4504,  0.5301, -1.1206, -0.5896],
        ...                        [ 0.7439,  0.4022,  0.5913,  0.1511],
        ...                        [-0.0523, -1.0082,  0.5536, -1.2748],
        ...                        [ 0.5151, -0.9396,  0.7223, -0.5500],
        ...                        [ 0.1083,  2.7311,  1.4429,  1.0640]])
        >>> targets = torch.tensor([3, 1, 1, 2, 0, 3, 0, 2]) # true class indices
        >>>
        >>> precision = CategoricalPrecision()
        >>> precision(inputs, targets)
        tensor(0.2500)

        if you're predicting categories for variable length sequences
        you can mask your targets with ``mask_target`` (-1 by default)
        predictions at those timesteps will then be excluded from the metric.

        >>> sequence_inputs = inputs.reshape(2, 4, 4) # (batch, timesteps, classes)
        >>> # note the last two timesteps in targets are masked with -1
        >>> masked_targets = torch.tensor([[ 3,  1,  1, 2],
        ...                                [ 0,  3, -1, -1]]) # (batch, timesteps)
        >>> precision(sequence_inputs, masked_targets)
        tensor(0.2500)
    """

    pass


@dataclass
class CategoricalFBeta(MaskingMixin, OneHotMixin, FBetaMixin, AverageMixin):
    """categorical fbeta score over possibly unnormalized scores given target indices.

    Built to have a somewhat similar interface to\
     `nn.CrossEntropyLoss <https://pytorch.org/docs/stable/data.html#torch.nn.CrossEntropyLoss>`_.

    Args:
        beta: beta value for weighting precision and recall.
            ``beta < 1`` weights precision higher, while ``beta > 1`` gives
            more weight to recall. Defaults to 1 (making the default the same as f1 score).
        mask_target: mask targets equal to this value. defaults to ``-1``.

    Example:
        >>> import torch
        >>> from hearth.metrics import CategoricalFBeta
        >>>
        >>> # here inputs is batch, feats and is not nessisarily normalized
        >>> inputs = torch.tensor([[-0.2956,  1.6050,  0.4113, -1.9041],
        ...                        [ 0.2095,  1.2959, -1.2466,  2.2302],
        ...                        [ 0.4702, -0.7506,  1.6751,  0.3370],
        ...                        [-0.4504,  0.5301, -1.1206, -0.5896],
        ...                        [ 0.7439,  0.4022,  0.5913,  0.1511],
        ...                        [-0.0523, -1.0082,  0.5536, -1.2748],
        ...                        [ 0.5151, -0.9396,  0.7223, -0.5500],
        ...                        [ 0.1083,  2.7311,  1.4429,  1.0640]])
        >>> targets = torch.tensor([3, 1, 1, 2, 0, 3, 0, 2]) # true class indices
        >>>
        >>> fbeta = CategoricalFBeta(beta=.5)
        >>> fbeta(inputs, targets)
        tensor(0.2083)

        if you're predicting categories for variable length sequences
        you can mask your targets with ``mask_target`` (-1 by default)
        predictions at those timesteps will then be excluded from the metric.

        >>> sequence_inputs = inputs.reshape(2, 4, 4) # (batch, timesteps, classes)
        >>> # note the last two timesteps in targets are masked with -1
        >>> masked_targets = torch.tensor([[ 3,  1,  1, 2],
        ...                                [ 0,  3, -1, -1]]) # (batch, timesteps)
        >>> fbeta(sequence_inputs, masked_targets)
        tensor(0.2500)
    """

    pass


@dataclass
class CategoricalF1(MaskingMixin, OneHotMixin, F1Mixin, AverageMixin):
    """categorical f1 score over possibly unnormalized scores given target indices.

    Built to have a somewhat similar interface to\
     `nn.CrossEntropyLoss <https://pytorch.org/docs/stable/data.html#torch.nn.CrossEntropyLoss>`_.

    Args:
        mask_target: mask targets equal to this value. defaults to ``-1``.

    Example:
        >>> import torch
        >>> from hearth.metrics import CategoricalF1
        >>>
        >>> # here inputs is batch, feats and is not nessisarily normalized
        >>> inputs = torch.tensor([[-0.2956,  1.6050,  0.4113, -1.9041],
        ...                        [ 0.2095,  1.2959, -1.2466,  2.2302],
        ...                        [ 0.4702, -0.7506,  1.6751,  0.3370],
        ...                        [-0.4504,  0.5301, -1.1206, -0.5896],
        ...                        [ 0.7439,  0.4022,  0.5913,  0.1511],
        ...                        [-0.0523, -1.0082,  0.5536, -1.2748],
        ...                        [ 0.5151, -0.9396,  0.7223, -0.5500],
        ...                        [ 0.1083,  2.7311,  1.4429,  1.0640]])
        >>> targets = torch.tensor([3, 1, 1, 2, 0, 3, 0, 2]) # true class indices
        >>>
        >>> f1 = CategoricalF1()
        >>> f1(inputs, targets)
        tensor(0.1667)

        if you're predicting categories for variable length sequences
        you can mask your targets with ``mask_target`` (-1 by default)
        predictions at those timesteps will then be excluded from the metric.

        >>> sequence_inputs = inputs.reshape(2, 4, 4) # (batch, timesteps, classes)
        >>> # note the last two timesteps in targets are masked with -1
        >>> masked_targets = torch.tensor([[ 3,  1,  1, 2],
        ...                                [ 0, 3, -1, -1]]) # (batch, timesteps)
        >>> f1(sequence_inputs, masked_targets)
        tensor(0.2500)
    """

    pass



from typing import Tuple, Union
import torch
from torch import Tensor
from dataclasses import dataclass
from hearth.metrics.base import Metric
from hearth.metrics.functional import accuracy, precision, recall, f1, fbeta
from hearth.metrics._utils import onehot_inputs_and_targets


@dataclass
class BinaryMixin(Metric):
    """mixin for binary metrics with logit and masking support.

    Args:
        mask_target: mask targets equal to this value. defaults to ``-1``.
        from_logits: if ``True`` inputs are expected to be unnormalized and a sigmoid
            function will be applied before comparison to targets. defaults to ``False``.
    """

    from_logits: bool = False

    def _prepare(self, inputs: Tensor, targets: Tensor) -> Tuple[Tensor, Tensor]:
        inputs, targets = super()._prepare(inputs, targets)
        targets = targets.squeeze(-1)
        inputs = inputs.reshape_as(targets)
        if self.from_logits:
            inputs = torch.sigmoid(inputs)
        return inputs, targets


@dataclass
class HardBinaryMixin(BinaryMixin):
    """This mixin rounds binary inputs."""

    def _prepare(self, inputs: Tensor, targets: Tensor) -> Tuple[Tensor, Tensor]:
        inputs, targets = super()._prepare(inputs, targets)
        return inputs.round(), targets


@dataclass
class MaskingMixin(Metric):
    """mixin for masking inputs and targets based on a flag value in the target.

    Args:
        mask_target: mask targets equal to this value. defaults to ``-1``.
    """

    mask_target: int = -1

    def _mask(self, inputs: Tensor, targets: Tensor) -> Tuple[Tensor, Tensor]:
        valid = targets != self.mask_target
        return inputs[valid], targets[valid]


class ArgmaxMixin(Metric):
    def _prepare(self, inputs: Tensor, targets: Tensor) -> Tuple[Tensor, Tensor]:
        indices = torch.argmax(inputs, dim=-1)
        return indices, targets


class OneHotMixin(Metric):
    def _prepare(self, inputs: Tensor, targets: Tensor) -> Tuple[Tensor, Tensor]:
        return onehot_inputs_and_targets(inputs, targets)


class AverageMixin(Metric):
    def _aggregate(self, result):
        return result.mean()


class AccuracyMixin(Metric):
    def forward(self, inputs: Tensor, targets: Tensor, **kwargs) -> Tensor:  # type: ignore
        return accuracy(inputs, targets)


@dataclass
class RecallMixin(Metric):

    eps: float = 1e-8
    dim: Union[Tuple[int], int] = 0

    def forward(self, inputs: Tensor, targets: Tensor, **kwargs) -> Tensor:  # type: ignore
        return recall(inputs, targets, eps=self.eps, dim=self.dim)


@dataclass
class PrecisionMixin(Metric):

    eps: float = 1e-8
    dim: Union[Tuple[int], int] = 0

    def forward(self, inputs: Tensor, targets: Tensor, **kwargs) -> Tensor:  # type: ignore
        return precision(inputs, targets, eps=self.eps, dim=self.dim)


@dataclass
class F1Mixin(Metric):

    eps: float = 1e-8
    dim: Union[Tuple[int], int] = 0

    def forward(self, inputs: Tensor, targets: Tensor, **kwargs) -> Tensor:  # type: ignore
        return f1(inputs, targets, eps=self.eps, dim=self.dim)


@dataclass
class FBetaMixin(Metric):

    beta: float = 1
    eps: float = 1e-8
    dim: Union[Tuple[int], int] = 0

    def forward(self, inputs: Tensor, targets: Tensor, **kwargs) -> Tensor:  # type: ignore
        return fbeta(inputs, targets, beta=self.beta, eps=self.eps, dim=self.dim)



from typing import Union, List

import torch
from torch import nn
from torch.nn import functional as F  # noqa: N812

from hearth.modules import BaseModule


class LayerNormSimple(BaseModule):
    """Layer norm without learnable bias and gain.

    This basically just wraps the standard torch LayerNorm so it has no elementwise
    affine by default.

    Args:
        normalized_shape: input shape from an expected input. If a single integer is used, it is
             treated as a singleton list, and this module will normalize over the last dimension
             which is expected to be of that specific size.
        eps: a value added to the denominator for numerical stability. Default: 1e-5

    **Reference**:
        `Xu et al: Understanding and Improving Layer Normalization
        <https://arxiv.org/abs/1911.07013>`_

    Example:
        >>> import torch
        >>> from hearth.modules import LayerNormSimple
        >>>
        >>> layer = LayerNormSimple(5) # 5 feats
        >>> layer
        LayerNormSimple(5, eps=1e-05)

        >>> x = torch.tensor([[-4.2721, -2.8159, -1.2351,  0.2388,  4.5915],
        ...                  [-0.9092, -3.9666, -1.4216, -4.7373,  2.0403],
        ...                  [-1.2210,  4.4796, -1.2772, -2.8781,  4.1868]])
        >>> layer(x)
        tensor([[-1.1730, -0.6950, -0.1761,  0.3077,  1.7365],
                [ 0.3694, -0.9000,  0.1566, -1.2200,  1.5940],
                [-0.6139,  1.2486, -0.6323, -1.1554,  1.1530]])

    """

    def __init__(self, normalized_shape: Union[int, List[int], torch.Size], eps: float = 1e-5):
        super().__init__()
        self.norm = nn.LayerNorm(normalized_shape, eps=eps, elementwise_affine=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.norm(x)

    def __repr__(self) -> str:
        shape_or_feats = self.norm.normalized_shape
        if len(shape_or_feats) == 1:  # type: ignore
            shape_or_feats = shape_or_feats[0]  # type: ignore
        return f'{self.__class__.__name__}({shape_or_feats}, eps={self.norm.eps})'


class MaskedLayerNorm(nn.LayerNorm):

    """layer norm that supports masked input.

    Only calcluates stats and normalizes non-padding timesteps.

    Inputs are expected to be of shape (B, T, ...) and mask is expected to be a boolean mask of
    of shape (B, T) where **valid** timesteps are True and invalid (padding) timesteps are false.

    Example:
        >>> from torch import nn
        >>> from hearth.modules import MaskedLayerNorm
        >>>
        >>> layer = MaskedLayerNorm(16)
        >>> mask = torch.tensor([[ True,  True,  True, False, False],
        ...                      [ True,  True,  True,  True,  True]])
        >>> inp = torch.rand(2, 5, 16)
        >>> out = layer(inp, mask)
        >>> out.shape
        torch.Size([2, 5, 16])

        >>> (out == 0.0).all(-1)
        tensor([[False, False, False,  True,  True],
               [False, False, False, False, False]])
    """

    def __init__(self, features: int, eps: float = 1e-5, elementwise_affine: bool = True):
        super().__init__(features, eps=eps, elementwise_affine=elementwise_affine)

    def forward(self, x: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:  # type: ignore
        """preform layer norm on valid timesteps in x as specifed by mask.

        Args:
            x: a tensor of shape (B, T, ...)
            mask: boolean mask of of shape (B, T) where **valid** timesteps are True and
                invalid (padding) timesteps are false.
        """
        out = torch.zeros_like(x)
        out[mask] += F.layer_norm(x[mask], self.normalized_shape, self.weight, self.bias, self.eps)
        return out



from typing import Mapping, TypeVar, Union, Dict
import re

T = TypeVar('T')

_TITLE_BOUNDRY = re.compile(r'(?=[A-Z])')


def to_snakecase(s: str):
    if len(s) > 1 and not s.islower():
        return f'{s[0]}{re.sub(_TITLE_BOUNDRY, "_", s[1:])}'.lower()
    return s


class MissingFromRegistryError(KeyError):
    pass


class Registry(Mapping[str, T]):
    """a mapping for key to class or function."""

    def __init__(self, name: str, **kwargs):
        self.name = name
        self._data: Dict[str, T] = {}
        for k, v in kwargs.items():
            self._register(k, v)

    def _normalize_key(self, k: str) -> str:
        return k.lower().strip()

    def _register(self, k: str, v: T):
        self._data[self._normalize_key(k)] = v

    def __getitem__(self, k: str) -> T:
        k = self._normalize_key(k)
        try:
            return self._data[k]
        except KeyError:
            raise MissingFromRegistryError(f'normalized key: {k} is not in {self.name} registry!')

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def register(self, key_or_obj: Union[str, T]):
        """decorator for registering a class or function."""
        # check if this was called with a
        # key arg or not...
        if isinstance(key_or_obj, str):
            key = key_or_obj
            # make a wrapper to extend as decorator

            def inner(decorated):
                self._register(key, decorated)
                return decorated

            return inner
        # otherwise its a decorated function or class...
        # register it with its name
        key = key_or_obj.__name__  # type: ignore
        self._register(key, key_or_obj)
        return key_or_obj



import sys

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal
from typing import Type, ClassVar, Optional, Union, Callable, List
from dataclasses import dataclass
from itertools import starmap
from functools import reduce, partial
import operator

from torch.optim.lr_scheduler import (
    _LRScheduler,
    LambdaLR,
    StepLR,
    MultiStepLR,
    ExponentialLR,
    CosineAnnealingLR,
    ReduceLROnPlateau,
)
from torch.optim.lr_scheduler import MultiplicativeLR  # type: ignore

from hearth.callbacks import Callback
from hearth.callbacks.utils import if_active


def _dotted_attrgetter(path: str, obj):
    return reduce(getattr, path.split('.'), obj)


class _LRSchedulerCallback(Callback):
    """Base class for lr scheduler callbacks.

    all lr scheduler callbacks and bases should inherit from this class.
    """

    _scheduler_cls: ClassVar[Type[_LRScheduler]] = NotImplemented

    start_epoch: int = 0
    end_epoch: Optional[int] = None

    def __init_subclass__(cls, scheduler_cls=None, *args, **kwargs):
        if scheduler_cls:
            cls._scheduler_cls = scheduler_cls
            # add start epoch and end epoch as optional
            # keyword only dataclass fields
            cls.__annotations__['start_epoch'] = int
            cls.__annotations__['end_epoch'] = Optional[int]
            cls.start_epoch = 0
            cls.end_epoch = None
            # autodataclass
            cls = dataclass(cls)
        return super().__init_subclass__(*args, **kwargs)

    def __post_init__(self):
        self.active = False

    def _exclude_from_scheduler_kwargs(self):
        return ('start_epoch', 'end_epoch')

    def _get_scheduler_kwargs(self):
        exclude = self._exclude_from_scheduler_kwargs()
        return {
            field: getattr(self, field)
            for field in self.__dataclass_fields__
            if field not in exclude
        }

    def _build_scheduler(self, optimizer):
        self.scheduler = self._scheduler_cls(optimizer, **self._get_scheduler_kwargs())

    def _on_activation(self, loop):
        self._build_scheduler(loop.optimizer)
        self.active = True

    def on_epoch_start(self, loop):
        if loop.epoch == self.start_epoch:
            self._on_activation(loop)

        elif loop.epoch == self.end_epoch:
            self.active = False


class EpochLRSchedulerCallback(_LRSchedulerCallback):
    """base class for lr scheduler callbacks that modify learning rate based on epoch."""

    @if_active
    def on_epoch_end(self, loop):
        self.scheduler.step()


class LambdaLRCallback(EpochLRSchedulerCallback, scheduler_cls=LambdaLR):
    """Sets the learning rate of each parameter group to the initial lr times a given function.

    Args:
        lr_lambda (function or list): A function which computes a multiplicative
            factor given an integer parameter epoch, or a list of such
            functions, one for each group in optimizer.param_groups.
        start_epoch: this callback will start at this epoch. Defaults to 0
        end_epoch: : this callback will stop at this epoch. Defaults to None

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from torch.utils.data import TensorDataset, DataLoader
        >>>
        >>> from hearth.loop import Loop
        >>> from hearth.callbacks import LambdaLRCallback
        >>>
        >>> # make fakey train data and model
        >>> x, y = torch.rand(130, 5), torch.rand(130, 1).round()
        >>> train = DataLoader(TensorDataset(x[:100], y[:100]), batch_size=16)
        >>> val = DataLoader(TensorDataset(x[-30:], y[-30:]), batch_size=16)
        >>> model = nn.Sequential(nn.Linear(5, 10), nn.ReLU(), nn.Linear(10, 1), nn.Sigmoid())
        >>>
        >>>
        >>> # make callback ... we'll stop early at epoch 3
        >>> callback = LambdaLRCallback(lambda x: x**2, end_epoch=3)
        >>> # setup the loop
        >>> loop = Loop(model,
        ...             optimizer=torch.optim.AdamW(model.parameters(), lr=0.1),
        ...             loss_fn = nn.BCELoss(),
        ...             callbacks = [callback]
        ...            )
        >>>
        >>> # run for 5 epochs
        >>> loop(train, val, epochs=5)
        >>> for row in loop.history:
        ...     print(row.epoch, row.lrs)
        0 {'group0': 0.1}
        1 {'group0': 0.1}
        2 {'group0': 0.4}
        3 {'group0': 0.9}
        4 {'group0': 0.9}
    """

    lr_lambda: Union[Callable, List[Callable]]


class MultiplicativeLRCallback(EpochLRSchedulerCallback, scheduler_cls=MultiplicativeLR):
    """Multiply the learning rate of each parameter group by the factor given in the specified \
    function.

    Args:
        lr_lambda (function or list): A function which computes a multiplicative
            factor given an integer parameter epoch, or a list of such
            functions, one for each group in optimizer.param_groups.
        start_epoch: this callback will start at this epoch. Defaults to 0
        end_epoch: : this callback will stop at this epoch. Defaults to None

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from torch.utils.data import TensorDataset, DataLoader
        >>>
        >>> from hearth.loop import Loop
        >>> from hearth.callbacks import MultiplicativeLRCallback
        >>>
        >>> # make fakey train data and model
        >>> x, y = torch.rand(130, 5), torch.rand(130, 1).round()
        >>> train = DataLoader(TensorDataset(x[:100], y[:100]), batch_size=16)
        >>> val = DataLoader(TensorDataset(x[-30:], y[-30:]), batch_size=16)
        >>> model = nn.Sequential(nn.Linear(5, 10), nn.ReLU(), nn.Linear(10, 1), nn.Sigmoid())
        >>>
        >>>
        >>> # make callback ... we'll start the scheduler rolling at epoch2
        >>> callback = MultiplicativeLRCallback(lambda x: .8, start_epoch=2)
        >>> # setup the loop
        >>> loop = Loop(model,
        ...             optimizer=torch.optim.AdamW(model.parameters(), lr=0.1),
        ...             loss_fn = nn.BCELoss(),
        ...             callbacks = [callback]
        ...            )
        >>>
        >>> # run for 6 epochs
        >>> loop(train, val, epochs=6)
        >>> for row in loop.history:
        ...     print(row.epoch, row.lrs)
        0 {'group0': 0.1}
        1 {'group0': 0.1}
        2 {'group0': 0.1}
        3 {'group0': 0.08}
        4 {'group0': 0.064}
        5 {'group0': 0.0512}
    """

    lr_lambda: Union[Callable, List[Callable]]


class StepLRCallback(EpochLRSchedulerCallback, scheduler_cls=StepLR):
    """Decays the learning rate of each parameter group by gamma every step_size epochs. Notice \
    that such decay can happen simultaneously with other changes to the learning rate from outside \
    this scheduler.

    Args:
        step_size (int): Period of learning rate decay.
        gamma (float): Multiplicative factor of learning rate decay. Defaults to 0.1.
        start_epoch: this callback will start at this epoch. Defaults to 0
        end_epoch: : this callback will stop at this epoch. Defaults to None

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from torch.utils.data import TensorDataset, DataLoader
        >>>
        >>> from hearth.loop import Loop
        >>> from hearth.callbacks import StepLRCallback
        >>>
        >>> # make fakey train data and model
        >>> x, y = torch.rand(130, 5), torch.rand(130, 1).round()
        >>> train = DataLoader(TensorDataset(x[:100], y[:100]), batch_size=16)
        >>> val = DataLoader(TensorDataset(x[-30:], y[-30:]), batch_size=16)
        >>> model = nn.Sequential(nn.Linear(5, 10), nn.ReLU(), nn.Linear(10, 1), nn.Sigmoid())
        >>>
        >>>
        >>> # make callback ...
        >>> callback = StepLRCallback(step_size=2, gamma=.9)
        >>> # setup the loop
        >>> loop = Loop(model,
        ...             optimizer=torch.optim.AdamW(model.parameters(), lr=0.1),
        ...             loss_fn = nn.BCELoss(),
        ...             callbacks = [callback]
        ...            )
        >>>
        >>> # run for 6 epochs
        >>> loop(train, val, epochs=6)
        >>> for row in loop.history:
        ...     print(row.epoch, row.lrs)
        0 {'group0': 0.1}
        1 {'group0': 0.1}
        2 {'group0': 0.09}
        3 {'group0': 0.09}
        4 {'group0': 0.081}
        5 {'group0': 0.081}
    """

    step_size: int
    gamma: float = 0.1


class MultiStepLRCallback(EpochLRSchedulerCallback, scheduler_cls=MultiStepLR):
    """Decays the learning rate of each parameter group by gamma once the number of epoch reaches \
    one of the milestones. Notice that such decay can happen simultaneously with other changes to \
    the learning rate from outside this scheduler.

    Args:
        milestones (list): List of epoch indices. Must be increasing. Also note that these values
            are epochs are from the perspective of the scheduler. so they will start when
            the scheduler starts not nessisarily based on the epoch in loop.
        gamma (float): Multiplicative factor of learning rate decay. Defaults to 0.1.
        start_epoch: this callback will start at this epoch. Defaults to 0
        end_epoch: : this callback will stop at this epoch. Defaults to None

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from torch.utils.data import TensorDataset, DataLoader
        >>>
        >>> from hearth.loop import Loop
        >>> from hearth.callbacks import MultiStepLRCallback
        >>>
        >>> # make fakey train data and model
        >>> x, y = torch.rand(130, 5), torch.rand(130, 1).round()
        >>> train = DataLoader(TensorDataset(x[:100], y[:100]), batch_size=16)
        >>> val = DataLoader(TensorDataset(x[-30:], y[-30:]), batch_size=16)
        >>> model = nn.Sequential(nn.Linear(5, 10), nn.ReLU(), nn.Linear(10, 1), nn.Sigmoid())
        >>>
        >>>
        >>> # make callback ...
        >>> callback = MultiStepLRCallback(milestones=[1, 3], gamma=.9)
        >>> # setup the loop
        >>> loop = Loop(model,
        ...             optimizer=torch.optim.AdamW(model.parameters(), lr=0.1),
        ...             loss_fn = nn.BCELoss(),
        ...             callbacks = [callback]
        ...            )
        >>>
        >>> # run for 6 epochs
        >>> loop(train, val, epochs=6)
        >>> for row in loop.history:
        ...     print(row.epoch, row.lrs)
        0 {'group0': 0.1}
        1 {'group0': 0.09}
        2 {'group0': 0.09}
        3 {'group0': 0.081}
        4 {'group0': 0.081}
        5 {'group0': 0.081}
    """

    milestones: List[int]
    gamma: float = 0.1

    def __post_init__(self):
        # check milestones is properly increasing
        if not all(starmap(operator.lt, zip(self.milestones[:-1], self.milestones[1:]))):
            raise ValueError('milestones must be an increasing list.')
        super().__post_init__()


class ExponentialLRCallback(EpochLRSchedulerCallback, scheduler_cls=ExponentialLR):
    """Decays the learning rate of each parameter group by gamma every epoch.

    Args:
        gamma (float): Multiplicative factor of learning rate decay. Defaults to 0.1.
        start_epoch: this callback will start at this epoch. Defaults to 0
        end_epoch: : this callback will stop at this epoch. Defaults to None.

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from torch.utils.data import TensorDataset, DataLoader
        >>>
        >>> from hearth.loop import Loop
        >>> from hearth.callbacks import ExponentialLRCallback
        >>>
        >>> # make fakey train data and model
        >>> x, y = torch.rand(130, 5), torch.rand(130, 1).round()
        >>> train = DataLoader(TensorDataset(x[:100], y[:100]), batch_size=16)
        >>> val = DataLoader(TensorDataset(x[-30:], y[-30:]), batch_size=16)
        >>> model = nn.Sequential(nn.Linear(5, 10), nn.ReLU(), nn.Linear(10, 1), nn.Sigmoid())
        >>>
        >>>
        >>> # make callback ...
        >>> callback = ExponentialLRCallback(gamma=.8)
        >>> # setup the loop
        >>> loop = Loop(model,
        ...             optimizer=torch.optim.AdamW(model.parameters(), lr=0.1),
        ...             loss_fn = nn.BCELoss(),
        ...             callbacks = [callback]
        ...            )
        >>>
        >>> # run for 5 epochs
        >>> loop(train, val, epochs=5)
        >>> for row in loop.history:
        ...     print(row.epoch, row.lrs)
        0 {'group0': 0.1}
        1 {'group0': 0.08}
        2 {'group0': 0.064}
        3 {'group0': 0.0512}
        4 {'group0': 0.04096}
    """

    gamma: float


class CosineAnnealingLRCallback(EpochLRSchedulerCallback, scheduler_cls=CosineAnnealingLR):
    """Set the learning rate of each parameter group using a cosine annealing schedule.

    Args:
        T_max: Maximum number of iterations.
        eta_min (float): Minimum learning rate. Default: 0.
        end_epoch: : this callback will stop at this epoch. Defaults to None.

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from torch.utils.data import TensorDataset, DataLoader
        >>>
        >>> from hearth.loop import Loop
        >>> from hearth.callbacks import CosineAnnealingLRCallback
        >>>
        >>> # make fakey train data and model
        >>> x, y = torch.rand(130, 5), torch.rand(130, 1).round()
        >>> train = DataLoader(TensorDataset(x[:100], y[:100]), batch_size=16)
        >>> val = DataLoader(TensorDataset(x[-30:], y[-30:]), batch_size=16)
        >>> model = nn.Sequential(nn.Linear(5, 10), nn.ReLU(), nn.Linear(10, 1), nn.Sigmoid())
        >>>
        >>>
        >>> # make callback ...
        >>> callback = CosineAnnealingLRCallback(T_max=4, eta_min=.0001)
        >>> # setup the loop
        >>> loop = Loop(model,
        ...             optimizer=torch.optim.AdamW(model.parameters(), lr=0.1),
        ...             loss_fn = nn.BCELoss(),
        ...             callbacks = [callback]
        ...            )
        >>>
        >>> # run for 6 epochs
        >>> loop(train, val, epochs=7)
        >>> for row in loop.history:
        ...     print(row.epoch, row.lrs)
        0 {'group0': 0.1}
        1 {'group0': 0.0853}
        2 {'group0': 0.0500}
        3 {'group0': 0.0147}
        4 {'group0': 0.0001}
        5 {'group0': 0.0147}
        6 {'group0': 0.0500}
    """

    T_max: int
    eta_min: float = 0.0


class ReduceLROnPlateauCallback(EpochLRSchedulerCallback, scheduler_cls=ReduceLROnPlateau):
    """Reduce learning rate when the metric specified by ``field`` on the loop has stopped \
    improving.

    Models often benefit from reducing the learning rate by a factor of 2-10 once learning \
    stagnates. This scheduler reads a metrics quantity and if no improvement is seen for a \
    'patience' number of epochs, the learning rate is reduced.

    Args:
        field: the name of the field to access on the loop, may be dotted path.
            Defaults to 'loss'.
        mode: if the metric is being minimized or maximized . Default: 'min'.
        factor: Factor by which the learning rate will be reduced. Default: 0.1.
        patience: number of stagnant epochs to wait before reducting lr. Defaults to 10.
        threshold: Threshold for measuring the new optimum to only focus on significant \
            changes. Default: 1e-4.
        threshold_mode: One of `rel`, `abs`. In `rel` mode,
            dynamic_threshold = best * ( 1 + threshold ) in 'max' mode or \
            best * ( 1 - threshold ) in `min` mode. In `abs` mode, \
            dynamic_threshold = best + threshold in `max` mode or best - threshold \
            in `min` mode. Default: 'rel'.
        cooldown: Number of epochs to wait before resuming normal operation after \
            lr has been reduced. Default: 0.
        min_lr: lower bound on the learning rate of all param groups or each group \
            respectively. Default: 0.
        eps (float): Minimal decay applied to lr. If the difference between new and old lr is \
            smaller than eps, the update is ignored. Default: 1e-8.
        start_epoch: this callback will start at this epoch. Defaults to 0
        end_epoch: : this callback will stop at this epoch. Defaults to None

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from torch.utils.data import TensorDataset, DataLoader
        >>>
        >>> from hearth.loop import Loop
        >>> from hearth.callbacks import ReduceLROnPlateauCallback
        >>> _ = torch.manual_seed(0)
        >>>
        >>> # make fakey train data and model
        >>> x, y = torch.rand(130, 5), torch.rand(130, 1).round()
        >>> train = DataLoader(TensorDataset(x[:100], y[:100]), batch_size=16)
        >>> val = DataLoader(TensorDataset(x[-30:], y[-30:]), batch_size=16)
        >>> model = nn.Sequential(nn.Linear(5, 10), nn.ReLU(), nn.Linear(10, 1), nn.Sigmoid())
        >>>
        >>>
        >>> # make callback ...
        >>> callback = ReduceLROnPlateauCallback(field='loss', patience=1)
        >>> # setup the loop
        >>> loop = Loop(model,
        ...             optimizer=torch.optim.AdamW(model.parameters(), lr=0.1),
        ...             loss_fn = nn.BCELoss(),
        ...             callbacks = [callback]
        ...            )
        >>>
        >>> # run for 6 epochs
        >>> loop(train, val, epochs=7)
        >>> for row in loop.history:
        ...     print(row.epoch, row.lrs)
        0 {'group0': 0.1}
        1 {'group0': 0.1}
        2 {'group0': 0.1}
        3 {'group0': 0.01}
        4 {'group0': 0.01}
        5 {'group0': 0.001}
        6 {'group0': 0.001}
    """

    field: str = 'loss'
    mode: Literal['min', 'max'] = 'min'
    factor: float = 0.1
    patience: int = 10
    threshold: float = 1e-4
    threshold_mode: Literal['rel', 'abs'] = 'rel'
    cooldown: int = 0
    min_lr: float = 0.0
    eps: float = 1e-8

    def __post_init__(self):
        super().__post_init__()
        self._get_value = partial(_dotted_attrgetter, self.field)

    def _exclude_from_scheduler_kwargs(self):
        return super()._exclude_from_scheduler_kwargs() + ('field',)

    @if_active
    def on_epoch_end(self, loop):
        self.scheduler.step(self._get_value(loop))



__all__ = [
    'Metric' 'BinaryAccuracy',
    'BinaryF1',
    'BinaryFBeta',
    'BinaryPrecision',
    'BinaryRecall',
    'SoftBinaryRecall',
    'SoftBinaryPrecision' 'CategoricalRecall',
    'CategoricalPrecision',
    'CategoricalFBeta',
    'CategoricalF1',
    'CategoricalAccuracy',
    'MetricStack',
    'MultiHeadMetric' 'Running',
]
from .metrics import (
    BinaryAccuracy,
    BinaryF1,
    BinaryFBeta,
    BinaryPrecision,
    BinaryRecall,
    SoftBinaryRecall,
    SoftBinaryPrecision,
    CategoricalRecall,
    CategoricalPrecision,
    CategoricalFBeta,
    CategoricalF1,
    CategoricalAccuracy,
)
from .wrappers import Running
from .base import Metric, MetricStack, MultiHeadMetric



from typing import Tuple
import torch
from torch import Tensor


def onehot_inputs_and_targets(inputs: Tensor, targets: Tensor) -> Tuple[Tensor, Tensor]:
    expanded_targets = targets.unsqueeze(-1)
    fill = inputs.new_ones(expanded_targets.shape)
    pred = torch.zeros_like(inputs).scatter_add(-1, inputs.argmax(dim=-1, keepdim=True), fill)
    tru = torch.zeros_like(inputs).scatter_add(-1, expanded_targets, fill)
    return pred, tru



import torch
from torch import nn
from hearth.modules.base import BaseModule


class Residual(BaseModule):
    """wraps a block in a residual connection :math:`y = block(x) + x`.

    Args:
        block: the module to wrap.

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from hearth.modules import Residual
        >>> _ = torch.manual_seed(0)
        >>>
        >>> res = Residual(nn.Linear(4, 4))
        >>>
        >>> x = torch.rand(2, 4) # (batch, feats)
        >>> res(x)
        tensor([[ 0.6371,  1.5493,  0.0031, -0.0379],
                [ 0.3584,  0.8512,  0.5208, -0.7607]], grad_fn=<AddBackward0>)
    """

    def __init__(self, block: nn.Module):
        super().__init__()
        self.block = block

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """forward padd for ``Residual`` wrapper."""
        y = self.block(x)
        return x + y


class ReZero(Residual):
    """Implements ReZero residual connection around a block with dropout (as in transformer\
     implementation).

    **Reference**:
        `Bachlechner et al: ReZero is All You Need: Fast Convergence at Large Depth
        <https://arxiv.org/abs/2003.04887>`_

    Example:
        >>> import torch
        >>> from hearth.modules import ReZero
        >>>
        >>> transformation = nn.Sequential(nn.Linear(10, 10),
        ...                                nn.ReLU(),
        ...                                nn.Dropout(.1),
        ...                                nn.Linear(10, 10)
        ...                                )
        >>> re_zero = ReZero(transformation, dropout=.1)
        >>>
        >>> x = torch.normal(0, 1, size=(5, 10))
        >>> y = re_zero(x)
        >>> y.shape
        torch.Size([5, 10])

        since ``re_zero``'s weight parameter has not actually been trained ``y`` its going to be
        equal to ``x`` and nothing from the transformation will be added to the input...
        as training goes on this should change.

        >>> (y == x).all()
        tensor(True)
    """

    def __init__(self, block: nn.Module, dropout: float = 0.0):
        super().__init__(block=block)
        self.dropout = nn.Dropout(dropout)
        self.res_weight = nn.Parameter(torch.tensor([0.0]))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """forward for ReZero."""
        y = self.block(x)
        y = y * self.res_weight
        return x + self.dropout(y)


class TimeMasked(BaseModule):
    """Call the wrapped layer such that only valid timesteps are passed to the underlying layer.

    Inputs are expected to be of shape (B, T, ...) and mask is expected to be a boolean mask of
    of shape (B, T) where **valid** timesteps are True and invalid (padding) timesteps are false
    underlying layer is expected to accept a single input of the masked shape.

    Example:
        >>> from torch import nn
        >>> from hearth.modules import TimeMasked
        >>>
        >>> layer = TimeMasked(nn.Sequential(nn.Linear(8, 12), nn.LayerNorm(12)))
        >>> mask = torch.tensor([[ True,  True,  True, False, False],
        ...                      [ True,  True,  True,  True,  True]])
        >>> inp = torch.rand(2, 5, 8)
        >>> out = layer(inp, mask)
        >>> out.shape
        torch.Size([2, 5, 12])

        >>> (out == 0.0).all(-1)
        tensor([[False, False, False,  True,  True],
               [False, False, False, False, False]])

    """

    def __init__(self, layer: nn.Module):
        super().__init__()
        self.layer = layer

    def blocks(self):
        if hasattr(self.layer, 'blocks'):
            yield from self.layer.blocks()
        yield self.layer

    def forward(self, x: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        """call the underlying layer using only timesteps from `mask` padding \
        invalid timesteps with 0.

        Args:
            x: a tensor of shape (B, T, ...)
            mask: boolean mask of of shape (B, T) where **valid** timesteps are True and
                invalid (padding) timesteps are false.
        """
        batch, timesteps = mask.shape
        masked_out = self.layer(x[mask])
        out = masked_out.new_zeros(batch, timesteps, *masked_out.shape[1:])
        out[mask] += masked_out
        return out



"""Samplers for sampling batch indices (or example indexes) for use with \
`torch.utils.data.DataLoader\
 <https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader>`_
"""
from typing import Sized, List, Iterator
import torch
from torch.utils.data import Sampler, DataLoader

from more_itertools import windowed


class SubsequenceSampler(Sampler):
    """a batch sampler that keeps sequences aligned within batches but shuffles batches.

    Often when training on a dataset that represents a full sequence we may want to
    window that sequence into multiple subsequences and shuffle those. Additionally
    when batch_size is not divisible by dataset length we choose a different short sequence
    on each iteration which means batches are not always starting and ending at the same
    position. The shortest batch will always be last to comply with expected behavior from
    other batchers.

    Args:
        dataset: something we can call ``len()`` on that represents the size of the dataset being
            batched. It can be the dataset itself, a tensor, list, range etc...
        batch_size: the desired batch size
        drop_shortest: if ``True`` drop the shortest batch on each iteration if the size of the
            dataset is not divisible by ``batch_size``. The short batch will be chosen randomly
            on each iteration (providing a little extra noise and ensuring we dont see exactly
            the same subsequences on each iteration). Defaults to False.

    Example:

        >>> import torch
        >>> from hearth.data.samplers import SubsequenceSampler
        >>> _ = torch.manual_seed(0)
        >>>
        >>> # subsequence sampler only dataset
        >>> # so it could be a range object or a Dataset, a tensor etc...
        >>> sampler = SubsequenceSampler(range(15), batch_size=4)
        >>> sampler
        SubsequenceSampler(dataset=range(0, 15), batch_size=4, drop_shortest=False)

        by default we will get one short batch of 3, since ``batch_size=4`` and  \
        ``drop_shortest=False``:

        >>> for batch in sampler:
        ...     print(batch)
        [11, 12, 13, 14]
        [7, 8, 9, 10]
        [3, 4, 5, 6]
        [0, 1, 2]


        between runs we shuffle and choose a new short batch:

        >>> for batch in sampler:
        ...     print(batch)
        [8, 9, 10, 11]
        [0, 1, 2, 3]
        [4, 5, 6, 7]
        [12, 13, 14]


        when ``drop_shortest=True`` the randomly chosen short batch will be dropped...

        >>> for batch in SubsequenceSampler(range(15), batch_size=4, drop_shortest=True):
        ...     print(batch)
        [11, 12, 13, 14]
        [0, 1, 2, 3]
        [7, 8, 9, 10]

        and again this will be different for each iteration:

        >>> for batch in SubsequenceSampler(range(15), batch_size=4, drop_shortest=True):
        ...     print(batch)
        [11, 12, 13, 14]
        [7, 8, 9, 10]
        [3, 4, 5, 6]


        when ``batch_size`` is divisible by the number of examples in the dataset
        behavior is a little more deterministic, since the indexes in each batch will always
        be the same:

        >>> for batch in SubsequenceSampler(range(12), batch_size=4):
        ...     print(batch)
        [0, 1, 2, 3]
        [4, 5, 6, 7]
        [8, 9, 10, 11]

        but batches will still be shuffled between iterations:

        >>> for batch in SubsequenceSampler(range(12), batch_size=4):
        ...     print(batch)
        [8, 9, 10, 11]
        [0, 1, 2, 3]
        [4, 5, 6, 7]

    """

    def __init__(self, dataset: Sized, batch_size: int, drop_shortest: bool = False):
        self.dataset = dataset
        self.batch_size = batch_size
        self.drop_shortest = drop_shortest

    def _get_lengths(self, total_samples: int) -> List[int]:
        short_batch_size = total_samples % self.batch_size
        n_batches = total_samples // self.batch_size + (short_batch_size > 0)
        lengths = [self.batch_size] * n_batches

        if short_batch_size != 0:
            short_batch_idx: int = torch.randint(high=n_batches, size=(1,)).item()  # type: ignore
            lengths[short_batch_idx] = short_batch_size

        return lengths

    def _batches(self) -> Iterator[List[int]]:
        total_samples = len(self.dataset)
        batches = torch.split_with_sizes(
            torch.arange(total_samples), self._get_lengths(total_samples)
        )
        sort_keys = torch.randperm(len(batches)).tolist()
        # here we ensure that the shortest batch is last
        yield from sorted([batches[i].tolist() for i in sort_keys], key=len, reverse=True)

    def __iter__(self) -> Iterator[List[int]]:
        for batch in self._batches():
            if not self.drop_shortest or (len(batch) == self.batch_size):
                yield batch

    def __len__(self) -> int:
        n = len(self.dataset)
        if self.drop_shortest:
            return n // self.batch_size  # type: ignore
        else:
            return (n + self.batch_size - 1) // self.batch_size

    def __repr__(self) -> str:
        name = self.__class__.__name__
        args = (
            f'dataset={self.dataset},'
            f' batch_size={self.batch_size},'
            f' drop_shortest={self.drop_shortest}'
        )
        return f'{name}({args})'


class BatchSubsequenceSampler(torch.utils.data.Sampler):
    """a batch sampler that generates nested ordered subsequence indexes aligned on batch.

    output indexes from this sampler will be  lists of shape [batch_size, sequence_lengh].

    Note:
        Internally this uses :class:`SubsequenceSampler` to generate batches and will always drop
        the short sequence if the length of the dataset is not divisible by `sequence_length`.

    Args:
        dataset: something we can call ``len()`` on that represents the size of the dataset being
            batched. It can be the dataset itself, a tensor, list, range etc...
        batch_size: the desired batch size
        sequence_length: desired sequence length
        drop_last: if ``True`` drop the batch if less than batch size.

    Example:
        >>> import torch
        >>> from hearth.data.samplers import SubsequenceSampler
        >>> _ = torch.manual_seed(0)
        >>> sampler = BatchSubsequenceSampler(range(113), batch_size=4, sequence_length=5)
        >>> len(sampler)
        6

        generate batch indexes for sequental sequences of shape (`batch_size`, `sequence_length`)
        the last batch may contain less sequences, but sequence lengths will never change.

        >>> for batch_idxes in sampler:
        ...    print(batch_idxes)
        [[20, 21, 22, 23, 24], [10, 11, 12, 13, 14], [55, 56, 57, 58, 59], [30, 31, 32, 33, 34]]
        [[5, 6, 7, 8, 9], [60, 61, 62, 63, 64], [80, 81, 82, 83, 84], [15, 16, 17, 18, 19]]
        [[0, 1, 2, 3, 4], [90, 91, 92, 93, 94], [103, 104, 105, 106, 107], [40, 41, 42, 43, 44]]
        [[85, 86, 87, 88, 89], [50, 51, 52, 53, 54], [35, 36, 37, 38, 39], [75, 76, 77, 78, 79]]
        [[45, 46, 47, 48, 49], [65, 66, 67, 68, 69], [70, 71, 72, 73, 74], [108, 109, 110, 111, 112]]
        [[98, 99, 100, 101, 102], [25, 26, 27, 28, 29]]
    """  # noqa : E501

    @classmethod
    def build_dataloader(
        cls,
        dataset: Sized,
        batch_size: int,
        sequence_length: int,
        drop_last: bool = False,
        **kwargs,
    ) -> DataLoader:
        """creates a new DataLoader using BatchSubsequenceSampler.

        Note:
            extra keyword arguments will be passed to \
            `torch.utils.data.DataLoader\
                <https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader>`_

        Args:
            dataset: actual dataset you'd like to use, this dataset should support nested
                multi-indexing, `torch.utils.dataset.TensorDataset` works just fine
            batch_size: the desired batch size for the sequences
            sequence_length: desired sequence length for subsequences
            drop_last: If true drop the last short batch. Defaults to True.
        """
        n_samples = len(dataset)
        # TODO: check dataset to ensure we can tolerate this kind of batch
        sampler = cls(
            range(n_samples),
            batch_size=batch_size,
            sequence_length=sequence_length,
            drop_last=drop_last,
        )
        return DataLoader(dataset, batch_sampler=sampler, **kwargs)  # type: ignore

    def __init__(
        self, dataset: Sized, batch_size: int, sequence_length: int, drop_last: bool = False
    ):
        self.subseq_sampler = SubsequenceSampler(
            dataset, batch_size=sequence_length, drop_shortest=True
        )
        self.batch_size = batch_size
        self.drop_last = drop_last

    def __len__(self) -> int:
        n = len(self.subseq_sampler)
        return (n // self.batch_size) + ((n % self.batch_size) > 0 and not self.drop_last)

    def __iter__(self) -> Iterator[List[List[int]]]:
        for sequences in windowed(self.subseq_sampler, self.batch_size, step=self.batch_size):
            seq_batch: List[List[int]] = list(filter(lambda x: x, sequences))  # type: ignore
            if len(seq_batch) == self.batch_size or not self.drop_last:
                yield seq_batch



from typing import Union, Sized
from dataclasses import dataclass
from torch.utils.data import Dataset, DataLoader
from torch.utils.data import TensorDataset as _TensorDataset

from hearth.containers import TensorDict
from hearth._collate import default_collate


class BatchesMixin:

    """mixin for supporting batches and collate method from datasets."""

    def collate(self, batch):
        return default_collate(batch)

    def batches(
        self,
        batch_size: int = 32,
        shuffle: bool = False,
        drop_last: bool = False,
        num_workers: int = 0,
        **kwargs,
    ) -> DataLoader:
        """return a Dataloader that iterates over batches of this dataset.

        Note:
            this method supports additional keyword args which will be passed to the dataloader.

        Args:
            batch_size: Defaults to 32.
            shuffle: if True shuffle the dataset. Defaults to False.
            drop_last: if true drop the last batch if it is less than batch size. Defaults to False.
            num_workers: number of workers. Defaults to 0.
        """
        return DataLoader(
            self,  # type: ignore
            batch_size=batch_size,
            num_workers=num_workers,
            drop_last=drop_last,
            shuffle=shuffle,
            collate_fn=self.collate,
            **kwargs,
        )


@dataclass
class XYDataset(Dataset, BatchesMixin):
    """basic dataset that returns a tuple of inputs and targets.

    supports :class:`hearth.containers.TensorDataset`
    """

    x: Union[Sized, TensorDict]
    y: Union[Sized, TensorDict]

    def __len__(self) -> int:
        if isinstance(self.x, TensorDict):
            return len(next(iter(self.x.values())))  # type: ignore
        return len(self.x)

    def __getitem__(self, index):
        return self.x[index], self.y[index]


class TensorDataset(_TensorDataset, BatchesMixin):
    pass



import os
import re
import sys

import matplotlib.pyplot as plt


from torchvision.io import read_image as _read_img, ImageReadMode

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal


ReadModeT = Literal['UNCHANGED', 'GRAY', 'GRAY_ALPHA', 'RGB', 'RGB_ALPHA']


_image_file_end_expr = re.compile(r'\.(png|jpe?g)$')


def is_valid_filename(path: str) -> bool:
    """check if a path to an image file is a valid png or jpg by checking filename.

    Args:
        path: path to check.

    Example:
        >>> from hearth.vision.utils import is_valid_filename
        >>>
        >>> is_valid_filename('blah.jpg')
        True

        >>> is_valid_filename('blah.png')
        True

        >>> is_valid_filename('blah.jpeg')
        True

        >>> is_valid_filename('nested/dir/blah.jpg')
        True

        >>> is_valid_filename('blah.txt')
        False
    """
    return _image_file_end_expr.search(path) is not None


def read_image(path: str, mode: ReadModeT = 'RGB'):
    """read an image (with proper handling for user dir) in given mode.

    Args:
        path: path to image file
        mode: string name for read mode, lowercased values will be translated to upercase. Defaults
            to 'RGB'.
    """
    return _read_img(os.path.expanduser(path), getattr(ImageReadMode, mode.upper()))


def show_img(img):
    """plot a single image."""
    return plt.imshow(img.transpose(0, -1))



from dataclasses import dataclass

from hearth.callbacks import Callback
from hearth.modules import BaseModule
from hearth.optimizers import LazyOptimizer
from hearth.events import UnbottleEvent, UnbottlingComplete


@dataclass
class FineTuneCallback(Callback):
    """This callback gradually unbottles blocks and adds them to the optimizer decaying learning \
    rate by depth.

    Note:
        only models derived from :class:`hearth.modules.BaseModule` and optimizers derived
        from :class:`hearth.modules.LazyOptimizer` currently supported.

    Args:
        start_epoch: start unbottling blocks at this epoch.
        unbottle_every: number of epochs to wait between unbottles.
        decay: lr for each block will be base_lr/ (decay^depth). Defaults to 2.6 (as in ULMfit).
        max_depth: maximum depth (in blocks) to unbottle. If -1 then keep unbottling until
            all blocks are trainable. Defaults to -1.

    **Active On:**
        - registration
        - epoch_start

    **Events Emitted:**
        - :class:`hearth.events.UnbottleEvent`

    **Accesses Loop Attributes:**
        - model
        - optimizer
    """

    start_epoch: int
    unbottle_every: int = 1
    decay: float = 2.6
    max_depth: int = -1

    def __post_init__(self):
        self._unbottling_complete: bool = False
        self._depth = 0

    def on_registration(self, loop):
        # check model and optimizer type
        if not isinstance(loop.model, BaseModule):
            raise TypeError(
                f'{self.__class__.__name__} only supports hearth.BaseModule subclasses.'
            )
        if not isinstance(loop.optimizer, LazyOptimizer):
            raise TypeError(
                f'{self.__class__.__name__} only supports hearth.LazyOptimizer subclasses.'
            )

    def _should_unbottle(self, epoch: int) -> bool:
        if not self._unbottling_complete:
            return (
                epoch >= self.start_epoch and (self.start_epoch + epoch) % self.unbottle_every == 0
            )
        return False

    def _get_lr(self, optimizer):
        # get base lr
        lr = optimizer.get_lrs()[0]
        return lr / (self.decay ** self._depth)

    def _unbottle(self, loop):
        next_block = loop.model.unbottle()
        # if theres nothing left this will be None and
        # so unbottling is complete
        if not next_block:
            self._unbottling_complete = True
            return UnbottlingComplete()
        # otherwise unbottle
        self._depth += 1
        lr = self._get_lr(loop.optimizer)
        loop.optimizer.add_model(next_block, lr=lr)
        return UnbottleEvent(epoch=loop.epoch, block=next_block.__class__.__name__, lr=lr)

    def on_epoch_start(self, loop):
        if self._should_unbottle(loop.epoch):
            event = self._unbottle(loop)
            loop.fire(event)
            if (self.max_depth > -1) and (self._depth >= self.max_depth):
                self._unbottling_complete = True
                if not isinstance(event, UnbottlingComplete):
                    loop.fire(UnbottlingComplete())



from typing import Sequence, Type, Optional
from collections import deque
from hearth.callbacks import Callback
from dataclasses import dataclass
from hearth.events import MonitoringEvent, Stagnation, EarlyStop


@dataclass
class EarlyStopping(Callback):
    """This callback stops the training loop early if it recieves a monitoring event for enough \
    steps.

    Args:
        patience: wait for this many steps (epochs) before triggering stopping.
        event_types: Types of monitoring events to checkpoint on.
            Default is (:class:`hearth.events.Stagnation`, )
        prepare_model: Optional callable which accepts and returns
             a :class:`BaseModule` for preparing the model to be saved.
             This function will always recieve a **copy** of the model on the loop just for
             safety. Defaults to None.
        field: If provided only stop on events where field matches this field.
        stage: If provided only stop on events where stage matches this stage.

    **Active On:**
        - event
        - epoch_end

    **Events Listened For:**
        - :class:`hearth.events.Stagnation` (default)

    **Events Emitted:**
        - :class:`hearth.events.EarlyStop`

    **Accesses Loop Attributes:**
        - should_stop

    **Modifies Loop Attributes:**
        - should_stop

    **Accesses Event Attributes:**
        - field
        - stage
        - steps
    """

    patience: int = 5
    event_types: Sequence[Type[MonitoringEvent]] = (Stagnation,)
    field: Optional[str] = None
    stage: Optional[str] = None

    def __post_init__(self):
        self._pending_event = deque([], maxlen=1)

    def _should_stop(self, event):
        if isinstance(event, self.event_types):
            if self.field:
                if event.field != self.field:
                    return False
            if self.stage:
                if event.stage != self.stage:
                    return False
            return event.steps > self.patience
        return False

    def on_epoch_end(self, loop):
        if self._pending_event:
            event = self._pending_event.pop()
            loop.fire(event)

    def on_event(self, loop, event):
        if self._should_stop(event):
            loop.should_stop = True
            self._pending_event.append(EarlyStop(epoch=loop.epoch))



from typing import Sequence, Callable, Optional, Union
import torch
from torch import nn
from contextlib import nullcontext
from hearth.callbacks import Callback, CallbackManager
from hearth.metrics import Running
from hearth.callbacks import History
from hearth.metrics import MetricStack
from hearth.losses import MultiHeadLoss
from hearth.optimizers import LazyOptimizer


class Loop:

    """The simplest kind of loop for basic supervised learning.

    Note:
        If you have more custom things you'd like to to that cant be handled
        in callbacks it's recommended to subclass this and overide the  ``handle_batch`` method.
    """

    stages = ('train', 'val')

    def __init__(
        self,
        model: nn.Module,
        optimizer: LazyOptimizer,
        loss_fn: Callable,
        metrics: Optional[Union[Callable, MetricStack, Sequence[Callable]]] = None,
        callbacks: Sequence[Callback] = (),
        history: Optional[History] = None,
        device: Union[torch.device, str] = 'cpu',
    ):
        self.model = model
        self.loss_fn = loss_fn
        self.to(device)
        self.optimizer = optimizer

        self.metrics = metrics
        self.n_batches = 0
        self.batches_seen = 0
        self.stage = self.stages[0]
        self.should_stop = False
        self.history = history if history is not None else History()
        self.callbacks = CallbackManager(self.history, *callbacks)
        self.callbacks.on_registration(self)
        self.epoch = self.history.last_epoch + 1 if len(self.history) else 0

    @property
    def optimizer(self):
        return self._optimizer

    @optimizer.setter
    def optimizer(self, optimizer):
        if isinstance(optimizer, LazyOptimizer) and not optimizer.initialized:
            optimizer.add_model(self.model)
        self._optimizer = optimizer

    def to(self, device: Union[torch.device, str]):
        self.device = device
        self.model.to(self.device)
        self._loss_fn.to(self.device)

    @property
    def metrics(self):
        return self._metrics

    @metrics.setter
    def metrics(self, metrics: Optional[Union[Callable, MetricStack, Sequence[Callable]]] = None):
        if metrics:
            if not isinstance(metrics, MetricStack):
                if isinstance(metrics, dict):
                    metrics = MetricStack(**metrics)
                elif isinstance(metrics, Sequence):
                    metrics = MetricStack(*metrics)
                else:
                    metrics = MetricStack(metrics)

            self._metrics = Running(metrics)
            self._has_metrics = True
        else:
            self._metrics = Running(MetricStack())
            self._has_metrics = False

    @property
    def loss_fn(self):
        return self._loss_fn

    @loss_fn.setter
    def loss_fn(self, loss_fn=Callable):
        self._is_multihead_loss = isinstance(loss_fn, MultiHeadLoss)
        self._loss_agg_key = loss_fn.aggregate_key if self._is_multihead_loss else None
        self._loss_fn = Running(loss_fn)

    @property
    def loss(self):
        return self.loss_fn.average

    @property
    def metric(self):
        if self._has_metrics:
            return self.metrics.average
        return None

    def grad_context(self):
        if self.stage == 'train':
            return nullcontext()
        return torch.no_grad()

    def _requires_backward(self) -> bool:
        return self.stage == 'train'

    def optimizer_step(self):
        self.optimizer.step()

    def _optimizer_step(self):
        self.callbacks.on_step_start(self)
        self.optimizer_step()
        self.callbacks.on_step_end(self)

    def compute_loss(self, yhat, ytru, **kwargs):
        return self.loss_fn(yhat, ytru, **kwargs)

    def _compute_loss(self, yhat, ytru, **kwargs):
        self.callbacks.on_loss_start(self)
        loss = self.compute_loss(yhat, ytru, **kwargs)
        self.callbacks.on_loss_end(self)
        if self._is_multihead_loss:
            return loss[self._loss_agg_key]
        return loss

    def compute_metric(self, yhat, ytru, **kwargs):
        return self.metrics(yhat, ytru, **kwargs)

    def _compute_metric(self, yhat, ytru, **kwargs):
        with torch.no_grad():
            self.callbacks.on_metric_start(self)
            metric = self.compute_metric(yhat, ytru, **kwargs)
            self.callbacks.on_metric_end(self)
        return metric

    def _backward(self, loss, **kwargs):
        self.callbacks.on_backward_start(self)
        self.backward(loss)
        self.callbacks.on_backward_end(self)

    def backward(self, loss, **kwargs):
        loss.backward()

    def _forward(self, x, **kwargs):
        with self.grad_context():
            return self.forward(x, **kwargs)

    def forward(self, x, **kwargs):
        return self.model(x, **kwargs)

    def handle_batch(self, batch):
        # do forward pass and get loss
        self.optimizer.zero_grad()
        # unpack the batch... you can override this if your batch differs...
        x, y = batch
        x = x.to(self.device)
        y = y.to(self.device)

        y_hat = self._forward(x)
        loss = self._compute_loss(y_hat, y)
        if self._requires_backward():
            self._backward(loss)
            self._optimizer_step()
        if self._has_metrics:
            self._compute_metric(y_hat, y)

    def handle_batches(self, batches):
        self.n_batches = len(batches)
        self.batches_seen = 0
        for batch in batches:
            self.callbacks.on_batch_start(self)
            self.handle_batch(batch)
            self.batches_seen += 1
            self.callbacks.on_batch_end(self)

    def handle_stage(self, stage, batches):
        self.stage = stage
        if self.stage == 'train':
            self.model.train()
        else:
            self.model.eval()
        if self._has_metrics:
            self.metrics.reset()
        self.loss_fn.reset()
        self.callbacks.on_stage_start(self)
        self.handle_batches(batches)
        self.callbacks.on_stage_end(self)

    def fire(self, event):
        self.callbacks.on_event(self, event)

    def __call__(self, train, val, epochs: int = 1):
        self.should_stop = False
        for _ in range(epochs):
            self.callbacks.on_epoch_start(self)
            for stage, batches in zip(self.stages, (train, val)):
                self.handle_stage(stage, batches)
            self.callbacks.on_epoch_end(self)
            self.epoch += 1
            if self.should_stop:
                break

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}(model={self.model},'
            f' optimizer={self.optimizer}'
            f' loss_fn={self.loss_fn!r}'
            f' metrics={self.metrics!r}'
            f' callbacks={self.callbacks})'
        )



import json
import pathlib


def save_json(obj, path: str):
    with open(path, 'w') as f:
        json.dump(obj, f)


def load_json(path: str):
    with open(path, 'r') as f:
        obj = json.load(f)
    return obj


def mkdirs_if_not_exist(path, verbose: bool = False):
    path = pathlib.Path(path)
    if not path.exists():
        if verbose:
            print(f'directory {path} does not exist. creating...')
        path.mkdir(parents=True, exist_ok=True)



from typing import ClassVar, Type
from copy import deepcopy
from torch import nn
from torch import optim
from hearth.grad import trainable_parameters, named_trainable_parameters
from more_itertools import partition


def _is_bias_namedparam(x):
    return x[0].endswith('bias')


def _has_weight_decay(defaults, **kwargs) -> bool:
    return bool(kwargs.get('weight_decay', 0) or defaults.get('weight_decay', 0))


class LazyOptimizer:
    """A base wrapper class for torch optimizers that is initilized without model parameters.

    Supports instantiation without model parameters, exclusion of biases from weight decay
    where applicable and adding only trainable parameters directly from models \
    (see :meth:`add_model`).
    """

    _optimizer_type: ClassVar[Type[optim.Optimizer]] = NotImplemented

    def __init__(self, *, decay_bias: bool = False, **kwargs):
        self.initialized = False
        self._optimizer_args = kwargs
        self.decay_bias = decay_bias

    def _init_optimizer(self, params):
        self.optimizer = self._optimizer_type(params, **self._optimizer_args)
        self.initialized = True

    def __getattr__(self, attr: str):
        if self.initialized:
            return getattr(self.optimizer, attr)
        try:
            return self._optimizer_args[attr]
        except KeyError:
            raise AttributeError(attr)

    def _params_from_model(self, model, **kwargs):
        return {'params': trainable_parameters(model), **kwargs}

    def _params_by_weight_decay(self, model, **kwargs):
        no_decay_kwargs = deepcopy(kwargs)
        no_decay_kwargs['weight_decay'] = 0.0
        named_weights, named_biases = partition(
            _is_bias_namedparam, named_trainable_parameters(model)
        )
        yield {'params': map(lambda x: x[-1], named_weights), **kwargs}
        if named_biases:
            yield {'params': map(lambda x: x[-1], named_biases), **no_decay_kwargs}

    def _init_optimizer_from_model(self, model):
        # if we dont use weight decay or are decaying biases same as weight
        # then just use trainable params from model
        if not _has_weight_decay(self._optimizer_args) or self.decay_bias:
            params = trainable_parameters(model)

        else:
            # otherwise get params by weight decay
            params = self._params_by_weight_decay(model)

        self._init_optimizer(params)

    def _update_from_model(self, model, **kwargs):
        # if we dont use weight decay or are decaying biases same as weight
        # then just use trainable params from model
        param_groups = []
        if not _has_weight_decay(self._optimizer_args, **kwargs) or self.decay_bias:
            param_groups.append(self._params_from_model(model, **kwargs))

        else:
            # otherwise get params by weight decay
            for g in self._params_by_weight_decay(model, **kwargs):
                param_groups.append(g)

        for group in param_groups:
            self.optimizer.add_param_group(group)

    def add_model(self, model: nn.Module, **kwargs):
        """add all trainable parameters from this model to the optimizer

        if extra kwargs are provided and the optimizer is already initilized these
        will be used to create param groups. If ::attr::`decay_bias` is False
        and :attr:`weight_decay` is non-zero, or a non-zero `weight_decay`
        value is passed to kwargs then weights and biases will be split into different
        parameter groups
        """
        if not self.initialized:
            self._init_optimizer_from_model(model)

        else:
            self._update_from_model(model, **kwargs)

    def __repr__(self) -> str:
        if self.initialized:
            return self.optimizer.__repr__()

        name = self.__class__.__name__
        kwargrepr = ', '.join(f'{k}={v!r}' for k, v in self._optimizer_args.items())
        return f'{name}({kwargrepr}, decay_bias={self.decay_bias})'

    def get_lrs(self):
        return [grp['lr'] for grp in self.param_groups]


class AdamW(LazyOptimizer):
    r"""Lazy version of AdamW.

    Supports instantiation without model parameters, exclusion of biases from weight decay
    where applicable and adding only trainable parameters directly from models
    (see :meth:`add_model`).
    The original Adam algorithm was proposed in `Adam: A Method for Stochastic Optimization`_.
    The AdamW variant was proposed in `Decoupled Weight Decay Regularization`_.

    Args:
        lr: learning rate (default: 1e-3)
        betas: coefficients used for computing running averages of gradient and its square
         (default: (0.9, 0.999))
        eps: term added to the denominator to improve numerical stability (default: 1e-8)
        weight_decay: weight decay coefficient (default: 1e-2)
        amsgrad: whether to use the AMSGrad variant of this algorithm from the paper
            `On the Convergence of Adam and Beyond`_  (default: False)
        decay_bias: if True include biases in weight_decay. (default: False)

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from hearth.optimizers import AdamW
        >>> from hearth.grad import freeze, unfreeze
        >>>
        >>> # no need to tell it about the model
        >>> opt = AdamW(lr=.1)
        >>> opt.initialized
        False

        define a model at some point, we'll also freeze
        a single layer:

        >>> model = nn.Sequential(nn.Linear(4, 5),
        ...                       nn.Linear(5, 6),
        ...                       nn.Linear(6, 7))
        >>> _=freeze(model[0])
        >>> opt.add_model(model)
        >>> opt.get_lrs()
        [0.1, 0.1]

        later we may unfreeze the first layer of our model
        and add that to our optimizer with a different learning
        rate:

        >>> _=unfreeze(model[0])
        >>> opt.add_model(model[0], lr=.01)
        >>> opt.get_lrs()
        [0.1, 0.1, 0.01, 0.01]

    .. _Adam\: A Method for Stochastic Optimization:
        https://arxiv.org/abs/1412.6980
    .. _Decoupled Weight Decay Regularization:
        https://arxiv.org/abs/1711.05101
    .. _On the Convergence of Adam and Beyond:
        https://openreview.net/forum?id=ryQu7f-RZ

    """

    _optimizer_type = optim.AdamW

    def __init__(
        self,
        lr=0.001,
        betas=(0.9, 0.999),
        eps=1e-08,
        weight_decay=0.01,
        amsgrad=False,
        decay_bias: bool = False,
    ):
        super().__init__(
            lr=lr,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            amsgrad=amsgrad,
            decay_bias=decay_bias,
        )


class Adadelta(LazyOptimizer):
    """Lazy version of Adadelta.

    Supports instantiation without model parameters, exclusion of biases from weight decay
    where applicable and adding only trainable parameters directly from models
    (see :meth:`add_model`).
    It has been proposed in `ADADELTA: An Adaptive Learning Rate Method`__.

    Args:
        rho: coefficient used for computing a running average of squared gradients (default: 0.9)
        eps: term added to the denominator to improve numerical stability (default: 1e-6)
        lr: coefficient that scale delta before it is applied to the parameters (default: 1.0)
        weight_decay: weight decay (L2 penalty) (default: 0) __ https://arxiv.org/abs/1212.5701
        decay_bias: if True include biases in weight_decay. (default: False)

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from hearth.optimizers import Adadelta
        >>> from hearth.grad import freeze, unfreeze
        >>>
        >>> # no need to tell it about the model
        >>> opt = Adadelta(lr=.1)
        >>> opt.initialized
        False

        define a model at some point, we'll also freeze a single layer:

        >>> model = nn.Sequential(nn.Linear(4, 5),
        ...                       nn.Linear(5, 6),
        ...                       nn.Linear(6, 7))
        >>> _ =freeze(model[0])
        >>> opt.add_model(model)
        >>> opt.get_lrs()
        [0.1]

        later we may unfreeze the first layer of our model and add that to our optimizer with a
        different learning rate:

        >>> _ = unfreeze(model[0])
        >>> opt.add_model(model[0], lr=.01)
        >>> opt.get_lrs()
        [0.1, 0.01]

    """

    _optimizer_type = optim.Adadelta

    def __init__(self, lr=1.0, rho=0.9, eps=1e-06, weight_decay=0, decay_bias: bool = False):
        super().__init__(lr=lr, rho=rho, eps=eps, weight_decay=weight_decay, decay_bias=decay_bias)


class Adagrad(LazyOptimizer):
    """Lazy version of Adagrad.

    Supports instantiation without model parameters, exclusion of biases from weight decay
    where applicable and adding only trainable parameters directly from models
    (see :meth:`add_model`).
    It has been proposed in `Adaptive Subgradient Methods for Online Learning
    and Stochastic Optimization`_.

    Args:
        lr: learning rate (default: 1e-2)
        lr_decay: learning rate decay (default: 0)
        weight_decay: weight decay (L2 penalty) (default: 0)
        eps: term added to the denominator to improve numerical stability (default: 1e-10)
        decay_bias: if True include biases in weight_decay. (default: False)

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from hearth.optimizers import Adagrad
        >>> from hearth.grad import freeze, unfreeze
        >>>
        >>> # no need to tell it about the model
        >>> opt = Adagrad(lr=.1)
        >>> opt.initialized
        False

        define a model at some point, we'll also freeze a single layer:

        >>> model = nn.Sequential(nn.Linear(4, 5),
        ...                       nn.Linear(5, 6),
        ...                       nn.Linear(6, 7))
        >>> _ =freeze(model[0])
        >>> opt.add_model(model)
        >>> opt.get_lrs()
        [0.1]

        later we may unfreeze the first layer of our model and add that to our optimizer with a
        different learning rate:

        >>> _ =unfreeze(model[0])
        >>> opt.add_model(model[0], lr=.01)
        >>> opt.get_lrs()
        [0.1, 0.01]

    .. _Adaptive Subgradient Methods for Online Learning and Stochastic
        Optimization: http://jmlr.org/papers/v12/duchi11a.html

    """

    _optimizer_type = optim.Adagrad

    def __init__(
        self,
        lr=0.01,
        lr_decay=0,
        weight_decay=0,
        initial_accumulator_value=0,
        eps=1e-10,
        decay_bias: bool = False,
    ):
        super().__init__(
            lr=lr,
            lr_decay=lr_decay,
            weight_decay=weight_decay,
            initial_accumulator_value=initial_accumulator_value,
            eps=eps,
            decay_bias=decay_bias,
        )


class Adam(LazyOptimizer):
    r"""Lazy version of Adam.

    Supports instantiation without model parameters, exclusion of biases from weight decay
    where applicable and adding only trainable parameters directly from models
    (see :meth:`add_model`).
    It has been proposed in `Adam: A Method for Stochastic Optimization`_.
    The implementation of the L2 penalty follows changes proposed in
    `Decoupled Weight Decay Regularization`_.

    Args:
        lr: learning rate (default: 1e-3)
        betas: coefficients used for computing running averages of gradient and its square
            (default: (0.9, 0.999))
        eps: term added to the denominator to improve numerical stability (default: 1e-8)
        weight_decay: weight decay (L2 penalty) (default: 0)
        amsgrad: whether to use the AMSGrad variant of this algorithm from the paper
            `On the Convergence of Adam and Beyond`_ (default: False)
            .. _Adam\: A Method for Stochastic Optimization:
        decay_bias: if True include biases in weight_decay. (default: False)

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from hearth.optimizers import Adam
        >>> from hearth.grad import freeze, unfreeze
        >>>
        >>> # no need to tell it about the model
        >>> opt = Adam(lr=.1)
        >>> opt.initialized
        False

        define a model at some point, we'll also freeze a single layer:

        >>> model = nn.Sequential(nn.Linear(4, 5),
        ...                       nn.Linear(5, 6),
        ...                       nn.Linear(6, 7))
        >>> _ = freeze(model[0])
        >>> opt.add_model(model)
        >>> opt.get_lrs()
        [0.1]

        later we may unfreeze the first layer of our model and add that to our optimizer with a
        different learning rate:

        >>> _ =unfreeze(model[0])
        >>> opt.add_model(model[0], lr=.01)
        >>> opt.get_lrs()
        [0.1, 0.01]

    .. _Adam\: A Method for Stochastic Optimization:
        https://arxiv.org/abs/1412.6980
    .. _Decoupled Weight Decay Regularization:
        https://arxiv.org/abs/1711.05101
    .. _On the Convergence of Adam and Beyond:
        https://openreview.net/forum?id=ryQu7f-RZ

    """

    _optimizer_type = optim.Adam

    def __init__(
        self,
        lr=0.001,
        betas=(0.9, 0.999),
        eps=1e-08,
        weight_decay=0,
        amsgrad=False,
        decay_bias: bool = False,
    ):
        super().__init__(
            lr=lr,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            amsgrad=amsgrad,
            decay_bias=decay_bias,
        )


class ASGD(LazyOptimizer):
    """Lazy version of ASGD.

    Supports instantiation without model parameters, exclusion of biases from weight decay
    where applicable and adding only trainable parameters directly from models
    (see :meth:`add_model`).
    It has been proposed in `Acceleration of stochastic approximation by averaging`_.

    Args:
        lr: learning rate (default: 1e-2)
        lambd: decay term (default: 1e-4)
        alpha: power for eta update (default: 0.75)
        t0: point at which to start averaging (default: 1e6)
        weight_decay: weight decay (L2 penalty) (default: 0)
            .. _Acceleration of stochasticapproximation by averaging:
        decay_bias: if True include biases in weight_decay. (default: False)

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from hearth.optimizers import ASGD
        >>> from hearth.grad import freeze, unfreeze
        >>>
        >>> # no need to tell it about the model
        >>> opt = ASGD(lr=.1)
        >>> opt.initialized
        False

        define a model at some point, we'll also freeze a single layer:

        >>> model = nn.Sequential(nn.Linear(4, 5),
        ...                       nn.Linear(5, 6),
        ...                       nn.Linear(6, 7))
        >>> _ =freeze(model[0])
        >>> opt.add_model(model)
        >>> opt.get_lrs()
        [0.1]

        later we may unfreeze the first layer of our model and add that to our optimizer with a
        different learning rate:

        >>> _ =unfreeze(model[0])
        >>> opt.add_model(model[0], lr=.01)
        >>> opt.get_lrs()
        [0.1, 0.01]

    .. _Acceleration of stochastic approximation by averaging:
        https://dl.acm.org/citation.cfm?id=131098

    """

    _optimizer_type = optim.ASGD

    def __init__(
        self,
        lr=0.01,
        lambd=0.0001,
        alpha=0.75,
        t0=1000000.0,
        weight_decay=0,
        decay_bias: bool = False,
    ):
        super().__init__(
            lr=lr, lambd=lambd, alpha=alpha, t0=t0, weight_decay=weight_decay, decay_bias=decay_bias
        )


class RMSprop(LazyOptimizer):
    """Lazy version of RMSprop.

    Supports instantiation without model parameters, exclusion of biases from weight decay
    where applicable and adding only trainable parameters directly from models
    (see :meth:`add_model`).
    Proposed by G. Hinton in his
    `course <https://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf>`_.

    The centered version first appears in `Generating Sequences
    With Recurrent Neural Networks <https://arxiv.org/pdf/1308.0850v5.pdf>`_.

    The implementation here takes the square root of the gradient average before
    adding epsilon (note that TensorFlow interchanges these two operations). The effective
    learning rate is thus :math:`\alpha/(\\sqrt{v} + \\epsilon)` where :math:`\alpha`
    is the scheduled learning rate and :math:`v` is the weighted moving average
    of the squared gradient.

    Args:
        lr: learning rate (default: 1e-2)
        momentum: momentum factor (default: 0)
        alpha: smoothing constant (default: 0.99)
        eps: term added to the denominator to improve numerical stability (default: 1e-8)
        centered: if ``True``, compute the centered RMSProp, the gradient is normalized by an
            estimation of its variance (Defaults to None)
        weight_decay: weight decay (L2 penalty) (default: 0)
        decay_bias: if True include biases in weight_decay. (default: False)

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from hearth.optimizers import RMSprop
        >>> from hearth.grad import freeze, unfreeze
        >>>
        >>> # no need to tell it about the model
        >>> opt = RMSprop(lr=.1)
        >>> opt.initialized
        False

        define a model at some point, we'll also freeze a single layer:

        >>> model = nn.Sequential(nn.Linear(4, 5),
        ...                       nn.Linear(5, 6),
        ...                       nn.Linear(6, 7))
        >>> _ = freeze(model[0])
        >>> opt.add_model(model)
        >>> opt.get_lrs()
        [0.1]

        later we may unfreeze the first layer of our model and add that to our optimizer with a
        different learning rate:

        >>> _= unfreeze(model[0])
        >>> opt.add_model(model[0], lr=.01)
        >>> opt.get_lrs()
        [0.1, 0.01]
    """

    _optimizer_type = optim.RMSprop

    def __init__(
        self,
        lr=0.01,
        alpha=0.99,
        eps=1e-08,
        weight_decay=0,
        momentum=0,
        centered=False,
        decay_bias: bool = False,
    ):
        super().__init__(
            lr=lr,
            alpha=alpha,
            eps=eps,
            weight_decay=weight_decay,
            momentum=momentum,
            centered=centered,
            decay_bias=decay_bias,
        )


class SGD(LazyOptimizer):
    """Lazy version of SGD.

    Supports instantiation without model parameters, exclusion of biases from weight decay
    where applicable and adding only trainable parameters directly from models
    (see :meth:`add_model`).
    Nesterov momentum is based on the formula from
    `On the importance of initialization and momentum in deep learning`__.

    Args:
        lr: learning rate
        momentum: momentum factor (default: 0)
        weight_decay: weight decay (L2 penalty) (default: 0)
        dampening: dampening for momentum (default: 0)
        nesterov: enables Nesterov momentum (default: False)
        decay_bias: if True include biases in weight_decay. (default: False)

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from hearth.optimizers import SGD
        >>> from hearth.grad import freeze, unfreeze
        >>>
        >>> # no need to tell it about the model
        >>> opt = SGD(lr=.1)
        >>> opt.initialized
        False

        define a model at some point, we'll also freeze a single layer:

        >>> model = nn.Sequential(nn.Linear(4, 5),
        ...                       nn.Linear(5, 6),
        ...                       nn.Linear(6, 7))
        >>> _=freeze(model[0])
        >>> opt.add_model(model)
        >>> opt.get_lrs()
        [0.1]

        later we may unfreeze the first layer of our model and add that to our optimizer with a
        different learning rate:

        >>> _=unfreeze(model[0])
        >>> opt.add_model(model[0], lr=.01)
        >>> opt.get_lrs()
        [0.1, 0.01]
    """

    _optimizer_type = optim.SGD

    def __init__(
        self,
        lr: float,
        momentum=0,
        dampening=0,
        weight_decay=0,
        nesterov=False,
        decay_bias: bool = False,
    ):
        super().__init__(
            lr=lr,
            momentum=momentum,
            dampening=dampening,
            weight_decay=weight_decay,
            nesterov=nesterov,
            decay_bias=decay_bias,
        )



import torch
from typing import Callable, Mapping, Union


class Running:
    """wrapper for metrics and losses for tracking running averages over batches.

    Args:
        fn: a loss or metric function.

    Example:
        >>> import torch
        >>> from torch import nn
        >>> _ = torch.manual_seed(0)
        >>> from hearth.metrics import Running
        >>>
        >>> running_loss = Running(nn.BCELoss())
        >>> running_loss
        Running(BCELoss())

        >>> predictions = torch.rand(10, 1, requires_grad=True)
        >>> targets = (torch.rand(10, 1) > .5) *1.0
        >>> running_loss(predictions, targets) # this should have grad!
        tensor(0.5636, grad_fn=<BinaryCrossEntropyBackward0>)

        >>> running_loss.average
        0.5636

        generate some  new random inputs and run again this time with smaller batch:

        >>> predictions = torch.rand(6, 1, requires_grad=True)
        >>> targets = (torch.rand(6, 1) > .5) *1.0
        >>> running_loss(predictions, targets)
        tensor(1.0943, grad_fn=<BinaryCrossEntropyBackward0>)

        >>> running_loss.average
        0.7625999748706818

        call ``reset()`` to reset it:

        >>> running_loss.reset()
        >>> running_loss.average
        0.0

    """

    def __init__(self, fn: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]):
        self.fn = fn
        self._reduction = getattr(fn, 'reduction', 'mean')
        self.reset()

    def reset(self):
        """reset all counters and totals on this metric."""
        self._batches_seen = 0
        self._samples_seen = 0
        self._total = 0

    @property
    def average(self) -> float:
        return self._total / max(self._samples_seen, 1)

    def _update(self, result, n_samples: int):
        if self._reduction == 'mean':
            self._total += result * n_samples
        elif self._reduction == 'sum':
            self._total += result
        self._samples_seen += n_samples
        self._batches_seen += 1

    def _get_n_samples(self, y):
        if isinstance(y, torch.Tensor):
            return y.shape[0]
        elif isinstance(y, Mapping):
            return self._get_n_samples(next(iter(y.values())))

    def __call__(self, inp, targets, **kwargs):
        res = self.fn(inp, targets, **kwargs)
        self._update(res.item(), self._get_n_samples(targets))
        return res

    def __repr__(self):
        return f'{self.__class__.__name__}({self.fn!r})'

    def to(self, device: Union[torch.device, str]):
        if hasattr(self.fn, 'to'):
            return self.fn.to(device)  # type: ignore
        return self



import torch
from torch import nn

from hearth.activations import get_activation
from hearth.modules.base import BaseModule
from hearth.modules.normalization import MaskedLayerNorm
from hearth.modules.wrappers import TimeMasked
from hearth.modules.attention import SelfAttention


class Boom(BaseModule):
    """feedforward part of std transformer network sometimes affectionatly referred to as the\
    bOOm layer (since it expands and contracts).


    Args:
        in_features: number of input features.
        scale: scale for intermediate size (the OO in bOOm). Defaults to 4 (commonly used in
            bert archetectures.)
        activation: named activation. Defaults to 'gelu'.
        dropout: Dropout rate for intermediate activation. Defaults to 0.1.

    Example:
        >>> import torch
        >>> from hearth.modules import Boom
        >>>
        >>> layer = Boom(16, scale=4, activation='gelu', dropout=0.1)
        >>> inp = torch.rand(10, 16)
        >>> layer(inp).shape
        torch.Size([10, 16])
    """

    def __init__(
        self, in_features: int, scale: int = 4, activation: str = 'gelu', dropout: float = 0.1
    ):
        super().__init__()
        self.in_features = self.out_features = in_features
        self.hidden_feats = self.in_features * scale
        self.in_layer = nn.Linear(self.in_features, self.hidden_feats)
        self.activation = get_activation(activation)
        self.dropout = nn.Dropout(dropout)
        self.out_layer = nn.Linear(self.hidden_feats, self.out_features)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.in_layer(x)
        x = self.activation(x)
        x = self.dropout(x)
        return self.out_layer(x)


class TransformerEncoderBlock(BaseModule):
    """a single transformer encoder block.

    Args:
        features: number of features.
        n_heads: number of attention heads.
        dropout: general dropout rate used in feedforward part of network and between
            residual connections. Defaults to 0.1.
        attn_dropout: dropout for self attention. Defaults to 0.1.
        activation: named activation for feedforward part of network. Defaults to 'gelu'.
        boom_scale: scale for intermediate size in feedforward network. Defaults to 4 (commonly
            used in bert archetectures).
        drop_head If true drop attention from entire attention heads as dropout strategy otherwise
            drop timesteps from different heads. Defaults to False.
        pre_norm: If true use pre-normalization strategy which may require less careful
            lr scheduling etc... Defaults to False as in bert-based models.
        layer_norm_eps: epsilon for layer norms used throught the network. Defaults to 1e-12.

    Example:
        >>> import torch
        >>> from hearth.modules import TransformerEncoderBlock
        >>>
        >>> layer = TransformerEncoderBlock(16, n_heads=4, boom_scale=4)
        >>> mask = torch.tensor([[ True,  True,  True, False, False],
        ...                      [ True,  True,  True,  True,  True]])
        >>> inp = torch.rand(2, 5, 16)
        >>> layer(inp, mask).shape
        torch.Size([2, 5, 16])

        with pre-norm scheme:

        >>> layer = TransformerEncoderBlock(16, n_heads=4, boom_scale=4, pre_norm=True)
        >>> layer(inp, mask).shape
        torch.Size([2, 5, 16])
    """

    def __init__(
        self,
        features: int,
        n_heads: int,
        dropout: float = 0.1,
        attn_dropout: float = 0.1,
        activation: str = 'gelu',
        boom_scale: int = 4,
        drop_head: bool = False,
        pre_norm: bool = False,
        layer_norm_eps: float = 1e-12,
    ):
        super().__init__()
        self.in_features = self.out_features = features
        self.attn_norm = MaskedLayerNorm(features, eps=layer_norm_eps)
        self.attn = SelfAttention(
            features, n_heads=n_heads, dropout=attn_dropout, drop_head=drop_head
        )
        self.attn_output_drop = nn.Dropout(dropout)

        self.ff_norm = MaskedLayerNorm(features, eps=layer_norm_eps)
        self.ff = TimeMasked(
            Boom(features, scale=boom_scale, dropout=dropout, activation=activation)
        )
        self.ff_output_drop = nn.Dropout(dropout)

        self.pre_norm = pre_norm

    def _sa_block(self, x: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        return self.attn_output_drop(self.attn(x, mask))

    def _ff_block(self, x: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        return self.ff_output_drop(self.ff(x, mask))

    def forward(self, x: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        if self.pre_norm:
            x = x + self._sa_block(self.attn_norm(x, mask), mask)
            x = x + self._ff_block(self.ff_norm(x, mask), mask)
        else:
            x = self.attn_norm(x + self._sa_block(x, mask), mask)
            x = self.ff_norm(x + self._ff_block(x, mask), mask)

        return x


class TransformerEncoder(BaseModule):

    """a stack of transformer encoder layers

    Args:
        features: number of features.
        layers: number of layers.
        n_heads: number of attention heads.
        dropout: general dropout rate used in feedforward part of network and between
            residual connections. Defaults to 0.1.
        attn_dropout: dropout for self attention. Defaults to 0.1.
        activation: named activation for feedforward part of network. Defaults to 'gelu'.
        boom_scale: scale for intermediate size in feedforward network. Defaults to 4 (commonly
            used in bert archetectures).
        drop_head If true drop attention from entire attention heads as dropout strategy otherwise
            drop timesteps from different heads. Defaults to False.
        pre_norm: If true use pre-normalization strategy which may require less careful
            lr scheduling etc... Defaults to False as in bert-based models.
        layer_norm_eps: epsilon for layer norms used throught the network. Defaults to 1e-12.

    Example:
        >>> import torch
        >>> from hearth.modules import TransformerEncoder
        >>>
        >>> model = TransformerEncoder(16, layers=3, n_heads=4, boom_scale=4)
        >>> mask = torch.tensor([[ True,  True,  True, False, False],
        ...                      [ True,  True,  True,  True,  True]])
        >>> inp = torch.rand(2, 5, 16)
        >>> model(inp, mask).shape
        torch.Size([2, 5, 16])

        >>> model.depth()
        3

    """

    def __init__(
        self,
        features: int,
        layers: int,
        n_heads: int,
        dropout: float = 0.1,
        attn_dropout: float = 0.1,
        activation: str = 'gelu',
        boom_scale: int = 4,
        drop_head: bool = False,
        pre_norm: bool = False,
        layer_norm_eps: float = 1e-12,
    ):
        super().__init__()
        self.layers = nn.ModuleList(
            [
                TransformerEncoderBlock(
                    features=features,
                    n_heads=n_heads,
                    dropout=dropout,
                    attn_dropout=attn_dropout,
                    activation=activation,
                    boom_scale=boom_scale,
                    drop_head=drop_head,
                    pre_norm=pre_norm,
                    layer_norm_eps=layer_norm_eps,
                )
                for _ in range(layers)
            ]
        )

    def blocks(self):
        yield from self.layers

    def forward(self, x: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        for mod in self.layers:
            x = mod(x, mask)
        return x



from collectionish import NumAttyDict


class NumberDict(NumAttyDict):
    """just like
    `collectionish.NumAttyDict\
    <https://collectionish.readthedocs.io/en/stable/_autosummary/collectionish.NumAttyDict.html>`_ \
     but with a special format method for better handling in logging.
    """

    def __format__(self, *args, **kwargs):
        return " ".join(f'{k}: {format(v, *args, **kwargs)}' for k, v in self.items())


class TensorDict(NumAttyDict):
    """a very basic keyed attribute accessable container for tensors.

    Can handle basic python math operators since it comes from :class:`NumAttyDict`
    """

    def to(self, device):
        """move all tensors in this Tensordict to device"""
        for v in self.values():
            v.to(device)
        return self

    def _idx_tensors(self, idx):
        return self.__class__({k: v[idx] for k, v in self.items()})

    def __getitem__(self, idx_or_key):
        if isinstance(idx_or_key, str):
            return dict.__getitem__(self, idx_or_key)
        return self._idx_tensors(idx_or_key)

    def item(self) -> 'NumAttyDict':
        """get python numbers out of this Tensordict by calling item on all tensors in it.

        Note:
            this will only work if all values in this tensordict are scalar tensors!

        Returns:
            :class:`NumberDict`
        """
        return NumberDict({k: v.item() for k, v in self.items()})  # type: ignore



import random
from torch import Tensor


def square_center_crop(img: Tensor, noise: float = 0.0) -> Tensor:
    """crop a single image of shape (channels, width, height) around the center with optional\
     random shift along the longest dimension.

    Args:
        img: single image tensor of shape (channels, width, height).
        noise: for shifting the crop window along the longest dimension Defaults to 0.0 (no noise).

    Example:
        >>> import torch
        >>> from hearth.vision.ops import square_center_crop
        >>>
        >>> img = torch.randint(0, 255, size=(3, 345, 256))
        >>> square_center_crop(img).shape
        torch.Size([3, 256, 256])

        >>> square_center_crop(img, noise=0.5).shape
        torch.Size([3, 256, 256])

        >>> square_center_crop(img.transpose(1, 2), noise=0.5).shape
        torch.Size([3, 256, 256])
    """
    _, w, h = img.shape
    # if height and width are same there's nothing to do...
    if w == h:
        return img
    out_size = min(w, h)
    cw = w - out_size
    ch = h - out_size
    noise = random.uniform(-noise, noise)
    # if theres's width to crop
    if cw:
        crop_left = int(round(cw / (2 + noise)))
        return img[:, crop_left : out_size + crop_left]
    # otherwise its longest on height...  crop that way...
    crop_top = int(round(ch / (2 + noise)))
    return img[:, :, crop_top : out_size + crop_top]



class _MultiHeadFunc:
    """base mixin for multiheaded losses and metrics"""

    def __init__(self, **kwargs):
        self._fns = {}
        for k, v in kwargs.items():
            if not callable(v):
                raise TypeError(
                    f'functions passed to {self.__class__.__name__} must be callable'
                    f' but got type {type(v)} for key {k}!'
                )
            self._fns[k] = v

    def items(self):
        return self._fns.items()

    def keys(self):
        return self._fns.keys()

    def values(self):
        return self._fns.values()

    def _argrepr(self):
        return ', '.join(f'{k}={v}' for k, v in self.items())

    def __repr__(self):
        return f'{self.__class__.__name__}({self._argrepr()})'



from typing import Iterator, Tuple
from torch import nn


def freeze(model: nn.Module) -> nn.Module:
    """freeze a model in place

    Args:
        model : model to freeze
    """
    for param in model.parameters():
        param.requires_grad = False
    return model


def unfreeze(model: nn.Module) -> nn.Module:
    """unfreeze a model in place

    Args:
        model : model to unfreeze
    """
    for param in model.parameters():
        param.requires_grad = True
    return model


def allgrad(model: nn.Module) -> bool:
    """return true if all parameters in the model require grad."""
    return all(p.requires_grad for p in model.parameters())


def anygrad(model: nn.Module) -> bool:
    """return true if any parameters in the model require grad."""
    return any(p.requires_grad for p in model.parameters())


def requires_grad(param: nn.Parameter) -> bool:
    """return true if the parameter requires grad"""
    return param.requires_grad


def trainable_parameters(model: nn.Module) -> Iterator[nn.Parameter]:
    """yields trainable parameters from a model."""
    yield from filter(requires_grad, model.parameters())


def named_trainable_parameters(model: nn.Module) -> Iterator[Tuple[str, nn.Parameter]]:
    """yields named trainable params from a model

    Args:
        model: model to get named params from

    Yields:
        tuples of name (str), param (nn.Parameter)
    """
    yield from filter(lambda x: requires_grad(x[-1]), model.named_parameters())  # type: ignore



from typing import Union
from dataclasses import dataclass
from torch import nn
from hearth.callbacks import Callback
from hearth.grad import trainable_parameters


@dataclass
class ClipGradNorm(Callback):
    """clips gradient norm of all trainable paramters of the ``loop.model`` on backward end.

    Args:
        max_norm: max norm of the gradients.
        norm_type: type of the used p-norm. Can be ``'inf'`` for infinity norm.
    """

    max_norm: float
    norm_type: Union[float, int] = 2

    def on_backward_end(self, loop):
        nn.utils.clip_grad_norm_(
            trainable_parameters(loop.model), max_norm=self.max_norm, norm_type=self.norm_type
        )


@dataclass
class ClipGradValue(Callback):
    """clips gradient of all trainable paramters of the ``loop.model`` on backward end.

    Args:
        clip_value: value to clip at.
    """

    clip_value: float

    def on_backward_end(self, loop):
        nn.utils.clip_grad_value_(trainable_parameters(loop.model), clip_value=self.clip_value)
